# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Functions used by the Toyplot implementation to validate inputs.  These are not intended for end-users.
"""


import numbers
import numpy

def instance(value, types):
    """Raise an exception if a value isn't one of the given type(s)."""
    if not isinstance(value, types):
        raise ValueError("Expected %s, received %s." % (types, type(value))) # pragma: no cover
    return value


def value_in(value, choices):
    """Raise an exception if a value doesn't match one of the given choices."""
    if value not in choices:
        raise ValueError("Expected one of %s, received %r." % (", ".join([repr(choice) for choice in choices]), value)) # pragma: no cover
    return value


def table_keys(table, keys, length=None, min_length=None, modulus=None):
    """Raise an exception if any of the given keys fails to match the column keys in the given table."""
    keys = string_vector(keys, length=length, min_length=min_length, modulus=modulus)
    allowed = list(table.keys())
    for key in keys:
        if key not in allowed:
            raise ValueError("Table key must match one of %s, received %s." % (", ".join(allowed), key)) # pragma: no cover
    return keys


def scalar(value):
    """Raise an exception if a value isn't a number."""
    if not isinstance(value, numbers.Number):
        raise ValueError("Expected a number, received %s." % value) # pragma: no cover
    return value


def scalar_array(value):
    """Raise an exception if a value isn't convertable to an array of numbers."""
    array = numpy.ma.array(value).astype("float64")
    array.mask = numpy.ma.mask_or(array.mask, ~numpy.isfinite(array))
    return array


def vector(value, length=None, min_length=None, modulus=None):
    """Raise an exception if a value isn't convertable to a 1D array."""
    array = numpy.ma.array(value)
    if array.shape == ():
        array = numpy.ma.repeat(array, 1)
    if array.ndim != 1:
        raise ValueError("Expected a vector.") # pragma: no cover
    if length is not None:
        if len(array) != length:
            raise ValueError("Expected %s values, received %s" % (length, len(array))) # pragma: no cover
    if min_length is not None:
        if len(array) < min_length:
            raise ValueError("Expected %s or more values, received %s" % (min_length, len(array))) # pragma: no cover
    if modulus is not None:
        if len(array) % modulus != 0:
            raise ValueError("Expected a multiple of %s values, received %s" % (modulus, len(array))) # pragma: no cover
    return array


def integer_vector(value, length=None, min_length=None, modulus=None):
    """Raise an exception if a value isn't convertable to a 1D array of integers."""
    return vector(value, length=length, min_length=min_length, modulus=modulus).astype("int64")


def scalar_vector(value, length=None, min_length=None, modulus=None):
    """Raise an exception if a value isn't convertable to a 1D array of numbers."""
    return vector(value, length=length, min_length=min_length, modulus=modulus).astype("float64")


def string_vector(value, length=None, min_length=None, modulus=None):
    """Raise an exception if a value isn't convertable to a 1D array of numbers."""
    return vector(value, length=length, min_length=min_length, modulus=modulus).astype("unicode")


def scalar_matrix(value, rows=None, columns=None):
    """Raise an exception if a value isn't convertable to a 2D array of numbers."""
    array = scalar_array(value)
    if array.ndim != 2:
        raise ValueError("Expected a matrix.") # pragma: no cover
    if rows is not None:
        if array.shape[0] != rows:
            raise ValueError("Expected %s rows, received %s." % (rows, array.shape[0])) # pragma: no cover
    if columns is not None:
        if array.shape[1] != columns:
            raise ValueError("Expected %s columns, received %s." % (columns, array.shape[1])) # pragma: no cover
    return array


def optional_string(value):
    """Raise an exception if a value isn't a string, or None."""
    if not isinstance(value, (str, type(None))):
        raise ValueError("Expected a string value or None, received %s." % value) # pragma: no cover
    return value


def filename(value):
    """Raise an exception if a value isn't a valid string filename, or None."""
    return optional_string(value)


def hyperlink(value):
    """Raise an exception if a value isn't a valid string hyperlink, or None."""
    return optional_string(value)



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Functionality for displaying a Toyplot canvas in a web browser."""


import collections

def show(canvases, title="Toyplot Figure"):
    """Display one or more canvases in a web browser.

    Uses Toyplot's preferred HTML+SVG+Javascript backend to display one-or-more
    interactive canvases in a web browser window.

    Because the canvases are displayed in a separate web browser process, this
    function returns immediately.

    Parameters
    ----------
    canvases: :class:`toyplot.canvas.Canvas` instance or sequence of :class:`toyplot.canvas.Canvas` instances.
      The canvases to be displayed.

    title: string, optional
      Optional page title to be displayed by the browser.
    """

    import os
    import tempfile
    import toyplot.canvas
    import toyplot.html
    import xml.etree.ElementTree as xml
    import webbrowser

    if not isinstance(canvases, (toyplot.canvas.Canvas, collections.abc.Iterable)):
        raise ValueError("Expected one or more instances of %s, received %s." % (toyplot.canvas.Canvas, type(canvases))) # pragma: no cover

    if isinstance(canvases, toyplot.canvas.Canvas):
        canvases = [canvases]

    html = xml.Element("html")
    head = xml.SubElement(html, "head")
    xml.SubElement(head, "title").text = title
    body = xml.SubElement(html, "body")
    for canvas in canvases:
        body.append(toyplot.html.render(canvas))

    fd, path = tempfile.mkstemp(suffix=".html")
    with os.fdopen(fd, "wb") as stream:
        stream.write(xml.tostring(html, method="html"))
    webbrowser.open("file://" + path, new=1, autoraise=True)



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Functionality for working with CSS style information."""


import numbers

import numpy

import toyplot.color

def require(css, allowed):
    """Validate that an object is usable as CSS style information.

    Parameters
    ----------
    css: dict or None
        The style dictionary to be validated.  An exception will be raised if it is
        not a valid style dictionary or None.
    allowed: sequence of strings
        The set of allowed style properties.  An exception will be raised if `css`
        contains any keys that aren't in this sequence.

    Returns
    -------
    style: dict
        The validated style dictionary.
    """
    if css is None:
        return css

    if not isinstance(css, dict):
        raise ValueError("Expected a dictionary of CSS styles or None, received %s." % css) # pragma: no cover

    for key, value in css.items():
        if key not in allowed:
            raise ValueError("Not an allowed CSS style: %s.  Use one of: %s" % (key, ", ".join(allowed))) # pragma: no cover
        if isinstance(value, numpy.ndarray) and value.dtype == toyplot.color.dtype:
            css[key] = toyplot.color.to_css(value)

    return css


class allowed(object):
    """Defines groups of allowable CSS property names."""

    #: Allowable CSS property names for filling areas.
    fill = set([
        "fill",
        "fill-opacity",
        "opacity",
        "stroke",
        "stroke-dasharray",
        "stroke-opacity",
        "stroke-width",
        ])

    #: Allowable CSS property names for stroking lines.
    line = set([
        "opacity",
        "stroke",
        "stroke-dasharray",
        "stroke-linecap",
        "stroke-opacity",
        "stroke-width",
        ])

    #: Allowable CSS property names for :ref:`markers`.
    marker = set([
        "fill",
        "fill-opacity",
        "opacity",
        "stroke",
        "stroke-opacity",
        "stroke-width",
        ])

    #: Allowable CSS property names for text.
    text = set([
        "alignment-baseline",
        "baseline-shift",
        "fill",
        "fill-opacity",
        "font-family",
        "font-size",
        "font-weight",
        "line-height",
        "opacity",
        "stroke",
        "stroke-opacity",
        "stroke-width",
        "text-anchor",
        "text-decoration-line",
        "text-shadow",
        "-toyplot-anchor-shift",
        "-toyplot-text-layout-box-visibility",
        "-toyplot-text-layout-line-visibility",
        "-toyplot-text-layout-visibility",
        "-toyplot-vertical-align",
        ])

def combine(*styles):
    """Combine multiple style specifications into one.

    Parameters
    ----------
    styles: sequence of :class:`dict` instances
        A collection of dicts containing CSS-compatible name-value pairs.

    Returns
    -------
    styles: :class:`dict` containing CSS-compatible name-value pairs.
    """
    computed_style = {}
    for style in styles:
        if style is not None:
            computed_style.update(style)
    return computed_style

def parse(css):
    """Parse a CSS style into a dict."""
    result = {}
    for declaration in css.split(";"):
        if declaration:
            key, value = declaration.split(":")
            result[key] = value.strip()
    return result

def _color_fixup(styles):
    """It turns-out that many applications and libraries (Inkscape, Adobe Illustrator, Qt)
    don't handle CSS rgba() colors correctly.  So convert them to CSS rgb colors and use
    fill-opacity / stroke-opacity instead."""

    if "fill" in styles:
        color = toyplot.color.css(styles["fill"])
        if color is not None:
            opacity = float(styles.get("fill-opacity", 1.0))
            styles["fill"] = "rgb(%.3g%%,%.3g%%,%.3g%%)" % (
                color["r"] * 100, color["g"] * 100, color["b"] * 100)
            styles["fill-opacity"] = str(color["a"] * opacity)
    if "stroke" in styles:
        color = toyplot.color.css(styles["stroke"])
        if color is not None:
            opacity = float(styles.get("stroke-opacity", 1.0))
            styles["stroke"] = "rgb(%.3g%%,%.3g%%,%.3g%%)" % (
                color["r"] * 100, color["g"] * 100, color["b"] * 100)
            styles["stroke-opacity"] = str(color["a"] * opacity)

    return styles

def to_css(*styles):
    """Convert one-or-more dicts containing CSS properties into a single CSS string."""
    declarations = []
    for key, value in sorted( _color_fixup(combine(*styles)).items()):
        if isinstance(value, numbers.Number):
            value = "{:.6g}".format(value)
        declarations.append("%s:%s" % (key, value))
    return ";".join(declarations)



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Functionality for managing markers (shapes used to highlight datums in plots and text).
"""


import copy
import xml.sax.saxutils

import numpy

import toyplot.style


class Marker(object):
    """Represents the complete specification of a marker's appearance."""
    def __init__(self, shape, mstyle, size, angle, label, lstyle):
        self._shape = shape
        self._mstyle = mstyle
        self._size = size
        self._angle = angle
        self._label = label
        self._lstyle = lstyle

    @property
    def shape(self):
        return self._shape

    @property
    def mstyle(self):
        return self._mstyle

    @property
    def size(self):
        return self._size

    @property
    def angle(self):
        return self._angle

    @property
    def label(self):
        return self._label

    @property
    def lstyle(self):
        return self._lstyle

    def __add__(self, other):
        if isinstance(other, str):
            return self.to_html() + other
        elif isinstance(other, toyplot.marker.Marker):
            result = copy.deepcopy(self)
            if other._shape is not None:
                result._shape = other._shape
            result._mstyle = toyplot.style.combine(result.mstyle, other._mstyle)
            if other._size is not None:
                result._size = other._size
            if other._angle is not None:
                result._angle = other._angle
            if other._label is not None:
                result._label = other._label
            result._lstyle = toyplot.style.combine(result.lstyle, other._lstyle)
            return result
        else:
            raise ValueError("Can't add toyplot.marker.Marker and %r" % other) # pragma: no cover

    def __eq__(self, other):
        return self._shape == other._shape and self._mstyle == other._mstyle and self._shape == other._shape and self._angle == other._angle and self._label == other._label and self._lstyle == other._lstyle

    def __hash__(self):
        return hash((self._shape, self._mstyle, self._size, self._angle, self._label, self._lstyle))

    def __radd__(self, other):
        return other + self.to_html()

    def __repr__(self):
        return self.to_html()

    def __format__(self, format_spec):
        return self.to_html()

    def to_html(self):
        """Convert a marker specification to HTML markup that can be embedded in rich text."""
        return """<marker%s%s%s%s%s%s/>""" % (
            " shape='%s'"% xml.sax.saxutils.escape(self._shape) if self._shape else "",
            " mstyle='%s'" % toyplot.style.to_css(self._mstyle) if self._mstyle else "",
            " size='%s'"% self._size if self._size else "",
            " angle='%s'" % self._angle if self._angle else "",
            " label='%s'" % xml.sax.saxutils.escape(self._label) if self._label else "",
            " lstyle='%s'" % toyplot.style.to_css(self._lstyle) if self._lstyle else "",
            )

    def intersect(self, p):
        """Compute the intersection between this marker's border and a line segment.

        Parameters
        ----------
        p: :class:`numpy.ndarray` with shape (2), required
            Relative coordinates of a line segment originating at the center of this marker.

        Returns
        -------
        dp: :class:`numpy.ndarray` with shape (2)
            Relative coordinates of the intersection with this marker's border.
        """
        if self._size:
            if self._shape in ["o", "oo", "o|", "o/", "o-", "o\\", "o+", "ox", "o*"]:
                p /= numpy.linalg.norm(p)
                p *= self._size / 2
                return p
            if self._shape in ["s"]:
                u = numpy.max(numpy.abs(p))
                p /= u
                p *= self._size / 2
                return p
            if self._shape and self._shape[0] == "r":
                width, height = self._shape[1:].split("x")
                width = float(width)
                height = float(height)

                ap = numpy.abs(p)
                if ap[1]:
                    if ap[0] / ap[1] > width / height:
                        p = p / ap[0] * self._size * width / 2
                    else:
                        p = p / ap[1] * self._size * height / 2
                else:
                    p = p / ap[0] * self._size * width / 2
                return p

        return numpy.zeros((2,))


def create(shape=None, mstyle=None, size=None, angle=None, label=None, lstyle=None):
    """Factory function for creating instances of :class:`toyplot.marker.Marker`."""
    return Marker(shape=shape, mstyle=mstyle, size=size, angle=angle, label=label, lstyle=lstyle)


def convert(value):
    """Construct an instance of :class:`toyplot.marker.Marker` from alternative representations."""
    if value is None:
        return value
    if isinstance(value, Marker):
        return value
    if isinstance(value, str):
        return Marker(shape=value, mstyle=None, size=None, angle=None, label=None, lstyle=None)
    raise ValueError("Can't convert %r to toyplot.marker.Marker." % value) # pragma: no cover


def from_html(html):
    """Convert a parsed xml.etree.ElementTree representation of a marker to a :class:`toyplot.marker.Marker` object."""
    size = html.get("size", None)
    if size is not None:
        size = float(size)

    angle = html.get("angle", None)
    if angle is not None:
        angle = float(angle)

    return Marker(
        shape=html.get("shape", None),
        mstyle=toyplot.style.parse(html.get("mstyle", "")),
        size=size,
        angle=angle,
        label=html.get("label", None),
        lstyle=toyplot.style.parse(html.get("lstyle", "")),
        )



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Strategy objects that compute the position and format of axis ticks and labels."""


import datetime

try:
    import arrow
except: # pragma: no cover
    pass
import custom_inherit
import numpy

import toyplot.broadcast


class TickLocator(object, metaclass=custom_inherit.DocInheritMeta(style="numpy_napoleon")):
    """Base class for tick locators - objects that compute the position and format of axis tick labels."""
    def ticks(self, domain_min, domain_max):
        """Return a set of ticks for the given domain.

        Parameters
        ----------
        domain_min, domain_max: number

        Returns
        -------
        locations : sequence of numbers
          Axis locations where ticks should be displayed.
        labels : sequence of strings
          Labels for each tick location.
        titles : sequence of strings
          Titles for each tick location.  Typically, backends render titles as tooltips.
        """
        raise NotImplementedError() # pragma: no cover


class Null(TickLocator):
    """Do-nothing tick locator that generates no ticks.

    Use :class:`toyplot.locator.Null` when you want to display an axis, but
    omit any ticks.
    """
    def ticks(self, domain_min, domain_max):
        return [], [], []


class Uniform(TickLocator):
    """Generate :math:`N` evenly spaced ticks that include the minimum and maximum values of a domain.

    Parameters
    ----------
    count: number, optional
      Number of ticks to generate.
    format : string, optional
      Format string used to generate labels from tick locations.
    """
    def __init__(self, count=5, format="{:g}"):
        self._count = count
        self._format = format

    def ticks(self, domain_min, domain_max):
        locations = numpy.linspace(
            domain_min, domain_max, self._count, endpoint=True)
        labels = [self._format.format(location) for location in locations]
        titles = numpy.repeat(None, len(labels))
        return locations, labels, titles


class Explicit(TickLocator):
    """Explicitly specify a collection of tick locations and labels for an axis.

    You must specify the set of locations, the set of labels, or both.  If you
    only specify locations, the labels will be generated using a formatting
    string.  If you only specify labels, the locations will default to integers
    in the range :math:`[0, n)`.  The latter behavior is especially useful for
    categorical axes.

    Parameters
    ----------
    locations : sequence numbers, optional
      Axis locations where ticks should be displayed.
    labels : sequence of strings, optional
      Labels for each tick location.
    titles : sequence of strings, optional
      Titles for each tick location.  Typically, backends render titles as tooltips.
    format : string, optional
      Format string used to generate labels from tick locations.
    """
    def __init__(
            self,
            locations=None,
            labels=None,
            titles=None,
            format="{:g}"):

        if locations is not None and labels is not None:
            locations = numpy.array(locations).astype("float64")
            labels = numpy.array(labels).astype("unicode")
            if len(locations) != len(labels):
                raise ValueError("Location and label counts must match.") # pragma: no cover
        elif locations is not None:
            locations = numpy.array(locations).astype("float64")
            labels = [format.format(location) for location in locations]
        elif labels is not None:
            labels = numpy.array(labels).astype("str")
            locations = numpy.arange(len(labels))
        else:
            raise ValueError("Must supply locations, labels, or both.") # pragma: no cover
        titles = toyplot.broadcast.pyobject(titles, len(locations))

        self._locations = locations
        self._labels = labels
        self._titles = titles

    def ticks(self, domain_min, domain_max):
        return self._locations, self._labels, self._titles

# An alpha version of the Talbot, Lin, Hanrahan tick mark generator for matplotlib.
# Described in "An Extension of Wilkinson's Algorithm for Positioning Tick Labels on Axes"
# by Justin Talbot, Sharon Lin, and Pat Hanrahan, InfoVis 2010.

# Implementation by Justin Talbot
# This implementation is in the public domain.
# Report bugs to jtalbot@stanford.edu

# A shortcoming:
# The weights used in the paper were designed for static plots where the extent of
# the tick marks unioned with the extent of the data defines the extent of the plot.
# In a plot where the extent of the plot is defined by the user (e.g. an interactive
# plot supporting panning and zooming), the weights don't work as well. In particular,
# you would want to retune them assuming that the tick labels must be inside
# the provided view range. You probably want higher weighting on simplicity and lower
# on coverage and possibly density. But I haven't experimented in any detail with this.
#
# If you do intend on using this for static plots in matplotlib, you should set
# only_inside to False in the call to Extended.extended. And then you should
# manually set your view extent to include the min and max ticks if they are outside
# the data range. This should produce the same results as the paper.


class Extended(TickLocator):
    """Generate ticks using "An Extension of Wilkinson's Algorithm for Positioning Tick Labels on Axes" by Talbot, Lin, and Hanrahan.

    Parameters
    ----------
    count: number, optional
      Desired number of ticks.  Note that the algorithm may produce fewer ticks
      than requested.
    steps: sequence of numbers, optional
      Prioritized list of "nice" values to use for generating ticks.
    only_inside: boolean
      If set to `True`, only ticks inside the axis domain will be generated.
    format : string, optional
      Format string used to generate labels from tick locations.
    """
    def __init__(self, count=5, steps=None, weights=None, only_inside=False, format="{0:.{digits}f}"):
        self._count = count
        self._steps = steps if steps is not None else [1, 5, 2, 2.5, 4, 3]
        self._weights = weights if weights is not None else [0.25, 0.2, 0.5, 0.05]
        self._only_inside = only_inside
        self._format = format

    def ticks(self, domain_min, domain_max):
        def coverage(dmin, dmax, lmin, lmax):
            range_ = dmax - dmin
            return 1 - 0.5 * (numpy.power(dmax - lmax,
                                          2) + numpy.power(dmin - lmin,
                                                           2)) / numpy.power(0.1 * range_,
                                                                             2)

        def coverage_max(dmin, dmax, span):
            range_ = dmax - dmin
            if span > range_:
                half = (span - range_) / 2.0
                return 1 - numpy.power(half, 2) / numpy.power(0.1 * range_, 2)
            else:
                return 1

        def density(k, m, dmin, dmax, lmin, lmax):
            r = (k - 1.0) / (lmax - lmin)
            rt = (m - 1.0) / (max(lmax, dmax) - min(lmin, dmin))
            return 2 - max(r / rt, rt / r)

        def density_max(k, m):
            if k >= m:
                return 2 - (k - 1.0) / (m - 1.0)
            else:
                return 1

        def simplicity(q, Q, j, lmin, lmax, lstep):
            eps = 1e-10
            n = len(Q)
            i = Q.index(q) + 1
            v = 1 if ((lmin % lstep < eps or (lstep - lmin % lstep)
                       < eps) and lmin <= 0 and lmax >= 0) else 0
            return (n - i) / (n - 1.0) + v - j

        def simplicity_max(q, Q, j):
            n = len(Q)
            i = Q.index(q) + 1
            v = 1
            return (n - i) / (n - 1.0) + v - j

        def legibility(lmin, lmax, lstep):
            return 1

        def legibility_max(lmin, lmax, lstep):
            return 1  # pragma: no cover

        def extended(dmin, dmax, m, Q, only_inside, w):
            n = len(Q)
            best_score = -2.0

            j = 1.0
            while j < float('infinity'):
                for q in Q:
                    sm = simplicity_max(q, Q, j)

                    if w[0] * sm + w[1] + w[2] + w[3] < best_score:
                        j = float('infinity')
                        break

                    k = 2.0
                    while k < float('infinity'):
                        dm = density_max(k, m)

                        if w[0] * sm + w[1] + w[2] * dm + w[3] < best_score:
                            break

                        delta = (dmax - dmin) / (k + 1.0) / j / q
                        z = numpy.ceil(numpy.log10(delta))

                        while z < float('infinity'):
                            step = j * q * numpy.power(10, z)
                            cm = coverage_max(dmin, dmax, step * (k - 1.0))

                            if w[0] * sm + w[1] * cm + \
                                    w[2] * dm + w[3] < best_score:
                                break

                            min_start = numpy.floor(
                                dmax / step) * j - (k - 1.0) * j
                            max_start = numpy.ceil(dmin / step) * j

                            if min_start > max_start:
                                z = z + 1
                                break

                            for start in range(
                                    int(min_start),
                                    int(max_start) + 1):
                                lmin = start * (step / j)
                                lmax = lmin + step * (k - 1.0)
                                lstep = step

                                s = simplicity(q, Q, j, lmin, lmax, lstep)
                                c = coverage(dmin, dmax, lmin, lmax)
                                d = density(k, m, dmin, dmax, lmin, lmax)
                                l = legibility(lmin, lmax, lstep)

                                score = w[0] * s + w[1] * \
                                    c + w[2] * d + w[3] * l

                                if score > best_score and (not only_inside or (lmin >= dmin and lmax <= dmax)):
                                    best_score = score
                                    best = (lmin, lmax, lstep, q, k)
                            z = z + 1
                        k = k + 1
                j = j + 1
            return best

        lmin, lmax, lstep, q, k = extended(
            domain_min, domain_max, self._count - 1, self._steps, self._only_inside, self._weights)
        locations = numpy.arange(k) * lstep + lmin
        digits = max(0, int(numpy.max(-numpy.floor(numpy.log10(lstep)), 0)))
        labels = [self._format.format(location, digits=digits) for location in locations]
        titles = numpy.repeat(None, len(labels))
        return locations, labels, titles


class Heckbert(TickLocator):
    """Generate ticks using the "Nice numbers for graph labels" algorithm by Paul Heckbert.

    Note that this locator can produce ticks outside the minimum / maximum axis
    domain.

    Parameters
    ----------
    count: number of ticks to generate
    format : string, optional
      Format string used to generate labels from tick locations.
    """
    def __init__(self, count=5, format="{0:.{digits}f}"):
        self._count = count
        self._format = format

    def ticks(self, domain_min, domain_max):
        def nicenum(x, rounding):
            exponent = numpy.floor(numpy.log10(x))
            fraction = x / numpy.power(10, exponent)
            if rounding:  # pragma: no cover
                if fraction < 1.5:
                    nice_fraction = 1
                elif fraction < 3:
                    nice_fraction = 2
                elif fraction < 7:
                    nice_fraction = 5
                else:
                    nice_fraction = 10
            else:  # pragma: no cover
                if fraction <= 1:
                    nice_fraction = 1
                elif fraction <= 2:
                    nice_fraction = 2
                elif fraction <= 5:
                    nice_fraction = 5
                else:
                    nice_fraction = 10
            return nice_fraction * numpy.power(10, exponent)

        tick_range = nicenum(domain_max - domain_min, rounding=False)
        tick_spacing = nicenum(tick_range / (self._count - 1), rounding=True)
        tick_min = numpy.floor(domain_min / tick_spacing) * tick_spacing
        tick_max = numpy.ceil(domain_max / tick_spacing) * tick_spacing
        locations = numpy.linspace(tick_min, tick_max, self._count)
        digits = max(0, int(numpy.max(-numpy.floor(numpy.log10(tick_spacing)), 0)))
        labels = [self._format.format(location, digits=digits) for location in locations]
        titles = numpy.repeat(None, len(labels))
        return locations, labels, titles


class Integer(TickLocator):
    """Generate evenly-spaced integer ticks

    Parameters
    ----------
    step: integer, optional
      Number of integer values to skip between labels.
    """
    def __init__(self, step=1):
        self._step = step

    def ticks(self, domain_min, domain_max):
        locations = numpy.arange(
            domain_min, domain_max + 1, self._step, dtype="int64")
        labels = [repr(location) for location in locations]
        titles = numpy.repeat(None, len(locations))
        return locations, labels, titles


class Log(TickLocator):
    """Generate ticks that are evenly spaced on a logarithmic scale

    Parameters
    ----------
    base : number, optional
      Logarithm base.
    format : string, optional
      Python format string, see
      `Format String Syntax <https://docs.python.org/2/library/string.html#format-string-syntax>`_
      for the syntax.  The format string will be called with a single positional
      argument containing the locator value, and two keyword arguments `base`
      and `exponent`.  If omitted, the locator will generate high-quality
      labels using superscript notation.

    Examples
    --------

    Generate locator labels using fixed decimal point notation and two significant digits:

    >>> locator = toyplot.locator.Log(format="{:.2g}")

    Generate locator labels using scientific notation and three significant digits:

    >>> locator = toyplot.locator.Log(format="{:.3e}")

    Generate locator labels using mixed scientific and decimal notation and four significant digits:

    >>> locator = toyplot.locator.Log(format="{:.4g}")

    Generate locator labels using "carat" notation:

    >>> locator = toyplot.locator.Log(format="{base}^{exponent}")

    """
    def __init__(self, base=10, format="{base}<sup> {exponent}</sup>"):
        self._base = base
        self._format = format

    def ticks(self, domain_min, domain_max):
        if domain_min < 0 and domain_max < 0:
            negative_exponents = numpy.arange(
                numpy.ceil(numpy.log10(numpy.abs(domain_min)) / numpy.log10(self._base)),
                numpy.floor(numpy.log10(numpy.abs(domain_max)) / numpy.log10(self._base)) - 1, -1)
        elif domain_min < 0:
            negative_exponents = numpy.arange(
                numpy.ceil(numpy.log10(numpy.abs(domain_min)) / numpy.log10(self._base)), -1, -1)
        else:
            negative_exponents = []
        negative_locations = -numpy.power(self._base, negative_exponents)

        if domain_max == 0 or domain_min == 0 or domain_min < 0 and domain_max > 0:
            linear_locations = [0]
        else:
            linear_locations = []

        if domain_min > 0 and domain_max > 0:
            positive_exponents = numpy.arange(
                numpy.floor(
                    numpy.log10(domain_min) /
                    numpy.log10(
                        self._base)),
                numpy.ceil(
                    numpy.log10(domain_max) /
                    numpy.log10(
                        self._base)) +
                1)
        elif domain_max > 0:
            positive_exponents = numpy.arange(
                0,
                numpy.ceil(
                    numpy.log10(domain_max) /
                    numpy.log10(
                        self._base)) +
                1) if domain_max != 0 else []
        else:
            positive_exponents = []
        positive_locations = numpy.power(self._base, positive_exponents)

        negative_labels = []
        for exponent, location in zip(negative_exponents, negative_locations):
            negative_labels.append(
                "-" + self._format.format(location, base=self._base, exponent=int(exponent))
                )

        linear_labels = [str(location) for location in linear_locations]

        positive_labels = []
        for exponent, location in zip(positive_exponents, positive_locations):
            positive_labels.append(
                self._format.format(location, base=self._base, exponent=int(exponent))
                )

        locations = numpy.concatenate(
            (negative_locations, linear_locations, positive_locations))
        labels = negative_labels + linear_labels + positive_labels
        titles = numpy.repeat(None, len(labels))
        return locations, labels, titles


def _year_generator(years):
    def implementation(domain_min, domain_max, timezone):
        start = domain_min.to(timezone)
        end = domain_max.to(timezone)
        for year in numpy.arange(start.year - (start.year % years), end.year + 1, years):
            value = arrow.Arrow(year=year, month=1, day=1, tzinfo=timezone)
            if value >= domain_min and value <= domain_max:
                yield value.timestamp()
    return implementation


def _month_generator(months):
    def implementation(domain_min, domain_max, timezone):
        start = domain_min.to(timezone)
        end = domain_max.to(timezone)
        for year in numpy.arange(start.year, end.year + 1):
            for month in numpy.arange(0, 12, months):
                value = arrow.Arrow(year=year, month=month + 1, day=1, tzinfo=timezone)
                if value >= domain_min and value <= domain_max:
                    yield value.timestamp()
    return implementation


def _day_generator(days):
    def implementation(domain_min, domain_max, timezone):
        start = domain_min.to(timezone)
        value = arrow.Arrow(year=start.year, month=start.month, day=start.day, tzinfo=timezone)
        while value <= domain_max:
            if value >= domain_min:
                yield value.timestamp()
            value += datetime.timedelta(days=days)
    return implementation


def _hour_generator(hours):
    def implementation(domain_min, domain_max, timezone):
        start = domain_min.to(timezone)
        value = arrow.Arrow(year=start.year, month=start.month, day=start.day, tzinfo=timezone)
        while value <= domain_max:
            if value >= domain_min:
                yield value.timestamp()
            value += datetime.timedelta(hours=hours)
    return implementation


def _minute_generator(minutes):
    def implementation(domain_min, domain_max, timezone):
        start = domain_min.to(timezone)
        value = arrow.Arrow(year=start.year, month=start.month, day=start.day, tzinfo=timezone)
        while value <= domain_max:
            if value >= domain_min:
                yield value.timestamp()
            value += datetime.timedelta(minutes=minutes)
    return implementation


def _second_generator(seconds):
    def implementation(domain_min, domain_max, timezone):
        start = domain_min.to(timezone)
        value = arrow.Arrow(year=start.year, month=start.month, day=start.day, tzinfo=timezone)
        while value <= domain_max:
            if value >= domain_min:
                yield value.timestamp()
            value += datetime.timedelta(seconds=seconds)
    return implementation


_intervals = [
    dict(duration=datetime.timedelta(days=365 * 1000), generator=_year_generator(1000), format="{0:YYYY}"),
    dict(duration=datetime.timedelta(days=365 * 100), generator=_year_generator(100), format="{0:YYYY}"),
    dict(duration=datetime.timedelta(days=365 * 10), generator=_year_generator(10), format="{0:YYYY}"),
    dict(duration=datetime.timedelta(days=365 * 5), generator=_year_generator(5), format="{0:YYYY}"),
    dict(duration=datetime.timedelta(days=365), generator=_year_generator(1), format="{0:YYYY}"),
    dict(duration=datetime.timedelta(days=90), generator=_month_generator(3), format="{0:MMMM} {0:YYYY}"),
    dict(duration=datetime.timedelta(days=30), generator=_month_generator(1), format="{0:MMMM} {0:YYYY}"),
    dict(duration=datetime.timedelta(days=7), generator=_day_generator(7), format="{0:ddd}, {0:MMM} {0:D}, {0:YYYY}"),
    dict(duration=datetime.timedelta(days=1), generator=_day_generator(1), format="{0:ddd}, {0:MMM} {0:D}"),
    dict(duration=datetime.timedelta(hours=12), generator=_hour_generator(12), format="{0:M}/{0:D} {0:HH}:{0:mm}"),
    dict(duration=datetime.timedelta(hours=6), generator=_hour_generator(6), format="{0:M}/{0:D} {0:HH}:{0:mm}"),
    dict(duration=datetime.timedelta(hours=4), generator=_hour_generator(4), format="{0:M}/{0:D} {0:HH}:{0:mm}"),
    dict(duration=datetime.timedelta(hours=1), generator=_hour_generator(1), format="{0:M}/{0:D} {0:HH}:{0:mm}"),
    dict(duration=datetime.timedelta(minutes=15), generator=_minute_generator(15), format="{0:M}/{0:D} {0:HH}:{0:mm}"),
    dict(duration=datetime.timedelta(minutes=10), generator=_minute_generator(10), format="{0:M}/{0:D} {0:HH}:{0:mm}"),
    dict(duration=datetime.timedelta(minutes=5), generator=_minute_generator(5), format="{0:M}/{0:D} {0:HH}:{0:mm}"),
    dict(duration=datetime.timedelta(minutes=1), generator=_minute_generator(1), format="{0:M}/{0:D} {0:HH}:{0:mm}"),
    dict(duration=datetime.timedelta(seconds=15), generator=_second_generator(15), format="{0:HH}:{0:mm}:{0:ss}"),
    dict(duration=datetime.timedelta(seconds=10), generator=_second_generator(10), format="{0:HH}:{0:mm}:{0:ss}"),
    dict(duration=datetime.timedelta(seconds=5), generator=_second_generator(5), format="{0:HH}:{0:mm}:{0:ss}"),
    dict(duration=datetime.timedelta(seconds=1), generator=_second_generator(1), format="{0:HH}:{0:mm}:{0:ss}"),
    ]


class Timestamp(TickLocator):
    """Generate ticks when the domain is UTC timestamps relative to the Unix epoch.

    Callers may explicitly specify the desired time interval as a string:

    >>> toyplot.locator.Timestamp(interval="hours")

    or as a tuple containing a quantity of units:

    >>> toyplot.locator.Timestamp(interval=(3, "days"))

    Instead of specifying an interval, the caller may supply the desired number
    of ticks, and the interval that produces the closest number of ticks will
    be used (note that the number of ticks produced may not match exactly):

    >>> toyplot.locator.Timestamp(count=4)

    If the count is not specified, it defaults to `7`.

    Parameters
    ----------
    count: number, optional
        Specifies the desired number of ticks.
    interval: string or (integer, string) tuple, optional
        Specifies the desired tick interval in anthropomorphic time units.
        Allowed units are "millenium", "millenia", "century", "centuries",
        "decade", "decades", "year", "years", "quarter", "quarters", "month",
        "months", "week", "weeks", "day", "days", "hour", "hours", "minute",
        "minutes", "second", and "seconds".
    timezone: string, optional
        Specifies a local timezone to be used for label generation.  Defaults
        to "utc".  Supports any timezone code allowed by :class:`arrow.arrow.Arrow`.
    format: string, optional
        Format string used to generate labels from tick locations.  The
        formatted value will be a :class:`arrow.arrow.Arrow` object, so any of the
        attributes and formatting provided by https://arrow.readthedocs.org may
        be used in the format.  For example, to display the full day of the
        week, month, day of the month without zero padding, and year, you could
        use:

    >>> toyplot.locator.Timestamp(format="{0:dddd}, {0:MMMM} {0:D}, {0:YYYY}")
    """
    def __init__(self, count=None, interval=None, timezone="utc", format=None):
        if interval is not None:
            if isinstance(interval, str):
                interval = (1, interval)
            interval = (int(interval[0]), interval[1].lower())

            amount, units = interval

            if units in ["millenium", "millenia"]:
                interval = dict(duration=datetime.timedelta(days=365 * 1000 * amount), generator=_year_generator(1000 * amount), format="{0:YYYY}")
            elif units in ["century", "centuries"]:
                interval = dict(duration=datetime.timedelta(days=365 * 100 * amount), generator=_year_generator(100 * amount), format="{0:YYYY}")
            elif units in ["decade", "decades"]:
                interval = dict(duration=datetime.timedelta(days=365 * 10 * amount), generator=_year_generator(10 * amount), format="{0:YYYY}")
            elif units in ["year", "years"]:
                interval = dict(duration=datetime.timedelta(days=365 * amount), generator=_year_generator(amount), format="{0:YYYY}")
            elif units in ["quarter", "quarters"]:
                interval = dict(duration=datetime.timedelta(days=90 * amount), generator=_month_generator(3 * amount), format="{0:MMMM} {0:YYYY}")
            elif units in ["month", "months"]:
                interval = dict(duration=datetime.timedelta(days=30 * amount), generator=_month_generator(amount), format="{0:MMMM} {0:YYYY}")
            elif units in ["week", "weeks"]:
                interval = dict(duration=datetime.timedelta(days=7 * amount), generator=_day_generator(7 * amount), format="{0:ddd}, {0:MMM} {0:D}, {0.year}")
            elif units in ["day", "days"]:
                interval = dict(duration=datetime.timedelta(days=amount), generator=_day_generator(amount), format="{0:ddd}, {0:MMM} {0:D}")
            elif units in ["hour", "hours"]:
                interval = dict(duration=datetime.timedelta(hours=amount), generator=_hour_generator(amount), format="{0:M}/{0:D} {0:HH}:{0:mm}")
            elif units in ["minute", "minutes"]:
                interval = dict(duration=datetime.timedelta(minutes=amount), generator=_minute_generator(amount), format="{0:M}/{0:D} {0:HH}:{0:mm}")
            elif units in ["second", "seconds"]:
                interval = dict(duration=datetime.timedelta(seconds=amount), generator=_second_generator(amount), format="{0:HH}:{0:mm}:{0:ss}")

        self._count = count
        self._interval = interval
        self._timezone = timezone
        self._format = format

    def ticks(self, domain_min, domain_max):
        interval = self._interval

        # If the caller didn't specify an interval, find one that yields the number of locations closest to the requested count
        if interval is None:
            if self._count is None:
                self._count = 7

            closest_difference = None

            for match in _intervals:
                count = (domain_max - domain_min) / match["duration"].total_seconds()
                difference = numpy.abs(self._count - count)

                if (closest_difference is None) or (difference < closest_difference):
                    interval = match
                    closest_difference = difference

        # Generate ticks
        generator = interval["generator"]
        label_format = interval["format"] if self._format is None else self._format

        locations = [location for location in generator(arrow.get(domain_min), arrow.get(domain_max), self._timezone)]
        labels = [label_format.format(arrow.get(location).to(self._timezone)) for location in locations]
        titles = numpy.repeat(None, len(labels))
        return locations, labels, titles



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Strategy objects for formatting text."""


import numbers
import warnings

import custom_inherit
import numpy

import toyplot


class Formatter(object, metaclass=custom_inherit.DocInheritMeta(style="numpy_napoleon")):
    """Abstract interface for formatters - objects that compute text representations from data."""

    def format(self, value):
        """Return a text representation of the given value.

        Parameters
        ----------
        value: value to be formatted

        Returns
        -------
        prefix : string
          Formatted data to be displayed before the separator.
        separator : string
          Separator between formatted data, or empty string.
        suffix : string
          Formatted data to be displayed after the separator, or empty string.
        """
        raise NotImplementedError() # pragma: no cover


class NullFormatter(Formatter):
    """Do-nothing formatter that returns empty strings."""
    def format(self, value):
        return "", "", ""


class BasicFormatter(Formatter):
    """Convert data to its default string representation."""
    def __init__(self, nanshow=True):
        self._nanshow = nanshow

    def format(self, value):
        try:
            if not self._nanshow and numpy.isnan(value):
                return "", "", ""
        except:
            pass

        return "%s" % value, "", ""


class FloatFormatter(BasicFormatter):
    """Formats floating-point values with aligned decimal points.

    Parameters
    ----------
    format: format string, optional
        Uses standard Python Format String Syntax.
    nanshow: bool, optional
        Set to `False` to hide NaN values.
    """
    def __init__(self, format="{:.6g}", nanshow=True):
        super(FloatFormatter, self).__init__(nanshow=nanshow)
        self._format = format

    def format(self, value):
        if isinstance(value, numbers.Number) and not numpy.isnan(value):
            formatted = self._format.format(value).split(".")
            if len(formatted) == 1:
                return formatted[0], "", ""
            return formatted[0], ".", formatted[1]

        return super(FloatFormatter, self).format(value)


class UnitFormatter(BasicFormatter):
    """Formats values with corresponding units

    Parameters
    ----------
    format: format string, optional
        Uses standard Python Format String Syntax.
    nanshow: bool, optional
        Set to `False` to hide NaN values.
    """

    def __init__(self, format="{:.6g}", nanshow=True, units='pt'):
        super(UnitFormatter, self).__init__(nanshow=nanshow)
        self._format = format
        self._units = units

        if units not in UnitFormatter._units:
            raise Exception("Incorrect type of units provided") #pragma: no cover

    def format(self, value):
        """Return a text representation of the given value.

        Parameters
        ----------
        value: value to be formatted
        units: units to be formatted

        Returns
        -------
        prefix : string
          Formatted data to be displayed before the separator.
        separator : string
          Separator between formatted data, or empty string.
        suffix : string
          Formatted data to be displayed after the separator, or empty string.
        units : string
          Formatted units to be displayed after the suffix, or empty string.
        """

        if isinstance(value, numbers.Number) and not numpy.isnan(value):
            formatted = self._format.format(value).split(".")
            if len(formatted) == 1:
                return formatted[0] +" " +UnitFormatter._units[self._units], "", ""
            return formatted[0], ".", formatted[1] +" " +UnitFormatter._units[self._units]

        return super(UnitFormatter, self).format(value)

UnitFormatter._units = {
    "centimeter": "cm",
    "centimeters": "cm",
    "cm": "cm",
    "decimeter": "dm",
    "decimeters": "dm",
    "dm": "dm",
    "in": "in",
    "inch": "in",
    "inches": "in",
    "m": "m",
    "meter": "m",
    "meters": "m",
    "millimeter": "mm",
    "millimeters": "mm",
    "mm": "mm",
    "pc": "pc",
    "pica": "pc",
    "picas": "pc",
    "pixel": "px",
    "pixels": "px",
    "point": "pt",
    "points": "pt",
    "pt": "pt",
    "px": "px",
}



class CurrencyFormatter(BasicFormatter):
    """Formats currency value with corresponding codes

    Parameters
    ----------
    places:  required number of places after the decimal point
    curr:    optional currency symbol before the sign (may be blank)
    sep:     optional grouping separator (comma, period, space, or blank)
    dp:      decimal point indicator (comma or period) only specify as blank when places is zero
    """
    def __init__(self, format="{:,.2f}", nanshow=True, curr='cad', sep=',', dp='.',):
        super(CurrencyFormatter, self).__init__(nanshow=nanshow)
        self._format = format
        self._curr = curr
        self._sep = sep
        self._dp = dp

        if self._curr not in CurrencyFormatter._codes:
            raise Exception("Incorrect currency provided") #pragma: no cover

    def format(self, value):
        """Return a text representation of the given currency value.

        Parameters
        ----------
        value: value to be formatted

        Returns
        -------
        codes : string
          Formatted currency code to be displayed before the prefix, or empty string.
        prefix : string
          Formatted data to be displayed before the separator.
        dp : string
          Separator between formatted data, or empty string.
        suffix : string
          Formatted data to be displayed after the separator, or empty string.
        """

        if isinstance(value, numbers.Number) and not numpy.isnan(value):
            formatted = self._format.format(value).split(".")
            return CurrencyFormatter._codes[self._curr] + formatted[0], self._dp, formatted[1]

        return super(CurrencyFormatter, self).format(value)

CurrencyFormatter._codes = {
        "aud": "$",
        "cad": "$",
        "eur": u"\u20AC",
        "gbp": u"\u00a3",
        "hkd": "HK$",
        "usd": "$",
}



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Functionality for managing scene graphs.

Note that the scenegraph is an internal implementation detail for developers
adding new functionality to Toyplot, casual Toyplot users will never need to
work with it.
"""

import collections


##########################################################################
# AdjacencyList

class AdjacencyList(object):
    """Adjaceny list representation for a directed graph."""
    def __init__(self):
        self._targets = collections.defaultdict(list)

    def add_edge(self, source, target):
        """Add a directed edge from `source` to `target`."""
        self._targets[source].append(target)

    def remove_edge(self, source, target):
        """Remove the directed edge from `source` to `target`, if it exists."""
        self._targets[source].remove(target)

    def sources(self, target):
        """Return nodes that are connected to `target` via incoming edges."""
        return [source for source, targets in self._targets.items() if target in targets]

    def targets(self, source):
        """Return nodes are connected to `source` via outgoing edges."""
        return self._targets.get(source, [])

##########################################################################
# SceneGraph

class SceneGraph(object):
    """Collection of graphs representing semantic relationships among a (small number) of canvas objects.
    """
    def __init__(self):
        self._relationships = {}

    def __repr__(self):
        result = "<toyplot.scenegraph.SceneGraph at 0x{:x}>".format(id(self))
        for relationship, graph in self._relationships.items():
            result += "\n  Relationship: {}".format(relationship)
            for source, targets in graph._targets.items():
                result += "\n    {!r} ->".format(source)
                for target in targets:
                    result += "\n      {!r}".format(target)
        return result

    def add_edge(self, source, relationship, target):
        """Add an edge of type `relationship` from `source` to `target`.
        """
        if not isinstance(relationship, str):
            raise ValueError("Relationship must be a string.")
        if relationship not in self._relationships:
            self._relationships[relationship] = AdjacencyList()
        self._relationships[relationship].add_edge(source, target)

    def remove_edge(self, source, relationship, target):
        """Remove an edge of type `relationship` from `source` to `target`, if one exists.
        """
        if not isinstance(relationship, str):
            raise ValueError("Relationship must be a string.")
        if relationship not in self._relationships:
            return
        self._relationships[relationship].remove_edge(source, target)

    def sources(self, relationship, target):
        """Return nodes that are connected to `target` via incoming edges of type `relationship`, if any."""
        if not isinstance(relationship, str):
            raise ValueError("Relationship must be a string.")
        if relationship not in self._relationships:
            return []
        return self._relationships[relationship].sources(target)

    def source(self, relationship, target):
        """Return a single node that is connected to `target` via an incoming edge of type `relationship`.

        Raises an exception if there isn't exactly one incoming edge of type `relationship`.
        """
        if not isinstance(relationship, str):
            raise ValueError("Relationship must be a string.")
        result = self.sources(relationship, target)
        if len(result) != 1:
            raise RuntimeError("Expected one source, found {}.".format(len(result)))
        return result[0]

    def targets(self, source, relationship):
        """Return nodes are connected to `source` via outgoing edges of type `relationship`."""
        if not isinstance(relationship, str):
            raise ValueError("Relationship must be a string.")
        if relationship not in self._relationships:
            return []
        return self._relationships[relationship].targets(source)




# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Functions for manipulating text."""


import copy
import sys
import xml.etree.ElementTree as xml

import numpy

import toyplot.font
import toyplot.require
import toyplot.style
import toyplot.transform
import toyplot.units


def extents(text, angle, style):
    """Compute canvas coordinate extents for a text string, based on the given angle and style.
    """
    text = toyplot.require.string_vector(text)
    angle = toyplot.require.scalar_vector(angle)
    style = toyplot.style.require(style, toyplot.style.allowed.text)

    # TODO: don't hard-code this.
    fonts = toyplot.font.ReportlabLibrary()

    layouts = numpy.array([toyplot.text.layout(string, style, fonts) for string in text])
    left = numpy.array([layout.left for layout in layouts])
    right = numpy.array([layout.right for layout in layouts])
    top = numpy.array([layout.top for layout in layouts])
    bottom = numpy.array([layout.bottom for layout in layouts])

    corner1 = numpy.column_stack((left, top))
    corner2 = numpy.column_stack((right, top))
    corner3 = numpy.column_stack((right, bottom))
    corner4 = numpy.column_stack((left, bottom))

    for index, theta in enumerate(angle):
        transformation = toyplot.transform.rotation(theta)
        corner1[index] = numpy.diag(corner1[index] * transformation)
        corner2[index] = numpy.diag(corner2[index] * transformation)
        corner3[index] = numpy.diag(corner3[index] * transformation)
        corner4[index] = numpy.diag(corner4[index] * transformation)

    left = numpy.minimum(corner1.T[0], numpy.minimum(
        corner2.T[0], numpy.minimum(corner3.T[0], corner4.T[0])))
    right = numpy.maximum(corner1.T[0], numpy.maximum(
        corner2.T[0], numpy.maximum(corner3.T[0], corner4.T[0])))
    top = -numpy.maximum(corner1.T[1], numpy.maximum(
        corner2.T[1], numpy.maximum(corner3.T[1], corner4.T[1])))
    bottom = -numpy.minimum(corner1.T[1], numpy.minimum(
        corner2.T[1], numpy.minimum(corner3.T[1], corner4.T[1])))

    return (left, right, top, bottom)


class Layout(object):
    """Top-level container for one-or-more lines of text."""
    def __init__(self, style):
        self.style = style
        self.children = []


class LineBox(object):
    """Container for one line of text."""
    def __init__(self, style):
        self.style = style
        self.children = []


class TextBox(object):
    """Container for a block of text that shares a single style."""
    def __init__(self, text, style):
        self.text = text
        self.style = style


class MarkerBox(object):
    """Container for a Toyplot marker specification."""
    def __init__(self, marker, style):
        self.marker = marker
        self.style = style


class PushHyperlink(object):
    """Marks the beginning of a hyperlink."""
    def __init__(self, href, target, style):
        self.href = href
        self.target = target
        self.style = style


class PopHyperlink(object):
    """Marks the end of a hyperlink."""
    def __init__(self, style):
        self.style = style


class _LineBreak(object):
    """Marks the position of a line break."""
    pass


def layout(text, style, fonts):
    """Convert text with markup into a layout that is ready to be rendered.

    Parameters
    ----------
    text: string, required
        Text with markup to be converted.
    style: dict, required
        Collection of CSS properties to be used as defaults for the text layout.
    fonts: :class:`toyplot.font.Library`, required
        Font library that will supply font information for the text layout.

    Returns
    -------
    layout: :class:`toyplot.text.Layout`
        The text layout contains a hierarchy of styled, positiioned nodes that
        are ready to be rendered.
    """
    def cascade_styles(style, node):
        """Cascades style information so that each node in an XML DOM has an explicit representation of each property/value pair."""
        if node.tag in ["b", "strong"]:
            style = toyplot.style.combine(style, {"font-weight": "bold"})
        elif node.tag in ["code"]:
            style = toyplot.style.combine(style, {"font-family": "monospace"})
        elif node.tag in ["em", "i"]:
            style = toyplot.style.combine(style, {"font-style": "italic"})
        elif node.tag in ["a"]:
            style = toyplot.style.combine(style, {"fill": "steelblue", "text-decoration-line": "none"})

        if "style" in node.attrib:
            node_style = toyplot.style.require(toyplot.style.parse(node.attrib["style"]), toyplot.style.allowed.text)
            style = toyplot.style.combine(style, node_style)

        node.set("style", copy.deepcopy(style))
        for child in node:
            cascade_styles(style, child)

    def compute_styles(reference_font_size, node):
        """Compute explicit numeric CSS pixel values for the baseline-shift, font-size, line-height, and -toyplot-text-anchor-shift properties."""
        font_size = node.get("style")["font-size"]
        font_size = toyplot.units.convert(font_size, target="px", default="px", reference=reference_font_size)

        baseline_shift = node.get("style")["baseline-shift"]
        baseline_shift = toyplot.units.convert(baseline_shift, target="px", default="px", reference=reference_font_size)

        toyplot_anchor_shift = node.get("style")["-toyplot-anchor-shift"]
        toyplot_anchor_shift = toyplot.units.convert(toyplot_anchor_shift, target="px", default="px", reference=reference_font_size)

        # Note that baseline shift is the opposite of canvas coordinates (positive values shift UP)
        if node.tag == "small":
            font_size *= 0.8
        elif node.tag == "sub":
            font_size *= 0.7
            baseline_shift -= 0.2 * font_size
        elif node.tag == "sup":
            font_size *= 0.7
            baseline_shift += 0.3 * font_size

        line_height = node.get("style")["line-height"]
        if line_height == "normal":
            line_height = "120%"
        line_height = toyplot.units.convert(line_height, target="px", default="px", reference=font_size)

        node.get("style")["baseline-shift"] = baseline_shift
        node.get("style")["font-size"] = "%spx" % font_size
        node.get("style")["line-height"] = line_height
        node.get("style")["-toyplot-anchor-shift"] = toyplot_anchor_shift

        for child in node:
            compute_styles(font_size, child)

    def build_formatting_model(node, root=None):
        """Convert the XML DOM into a flat layout containing text boxes and line breaks."""
        if node.tag == "body":
            root = Layout(node.get("style"))

        if node.tag in ["body", "b", "code", "i", "em", "small", "span", "strong", "sub", "sup"]:
            if node.text:
                root.children.append(TextBox(node.text, node.get("style")))
            for child in node:
                build_formatting_model(child, root)
                if child.tail:
                    root.children.append(TextBox(child.tail, node.get("style"))) # Note: the tail doesn't get the child's style
            return root

        if node.tag == "a":
            root.children.append(PushHyperlink(node.get("href"), node.get("target", None), node.get("style")))
            if node.text:
                root.children.append(TextBox(node.text, node.get("style")))
            for child in node:
                build_formatting_model(child, root)
                if child.tail:
                    root.children.append(TextBox(child.tail, node.get("style"))) # Note: the tail doesn't get the child's style
            root.children.append(PopHyperlink(node.get("style")))
            return root

        if node.tag == "marker":
            root.children.append(MarkerBox(toyplot.marker.from_html(node), node.get("style")))
            return root

        if node.tag == "br":
            root.children.append(_LineBreak())
            return root

        raise ValueError("Unknown tag: %s" % node.tag) # pragma: no cover

    def split_lines(layout):
        """Convert a flat layout into a two level hierarchy of line boxes containing text boxes."""
        children = []
        current_line = None

        for child in layout.children:
            if isinstance(child, _LineBreak):
                current_line = None
            else:
                if current_line is None:
                    current_line = LineBox(child.style)
                    children.append(current_line)
                current_line.children.append(child)

        layout.children = children

    def compute_size(fonts, layout):
        """Compute width + height for the layout + line boxes + text boxes."""
        for line in layout.children:
            for box in line.children:
                if isinstance(box, TextBox):
                    font = fonts.font(box.style)
                    box.width = font.width(box.text)

                    # Box baseline is the relative offset from the line baseline in canvas coordinates
                    alignment_baseline = box.style["alignment-baseline"]
                    if alignment_baseline == "alphabetic":
                        box.baseline = 0
                    elif alignment_baseline == "central":
                        box.baseline = font.ascent * 0.5
                    elif alignment_baseline == "hanging":
                        box.baseline = font.ascent
                    elif alignment_baseline == "middle":
                        box.baseline = font.ascent * 0.35
                    else:
                        raise ValueError("Unknown alignment-baseline value: %s" % alignment_baseline)

                    # Box top is the relative offset from the line baseline in canvas coordinates
                    box.top = box.baseline - font.ascent
                    # Box bottom is the relative offset from the line baseline in canvas coordinates
                    box.bottom = box.baseline - font.descent

                    box.height = box.bottom - box.top
                elif isinstance(box, MarkerBox):
                    font = fonts.font(box.style)

                    box.baseline = 0
                    box.top = box.baseline - font.ascent
                    box.bottom = box.baseline - font.descent
                    box.height = box.bottom - box.top
                    box.width = box.height
                elif isinstance(box, (PushHyperlink, PopHyperlink)):
                    box.baseline = 0
                    box.top = box.baseline
                    box.bottom = box.baseline
                    box.height = box.bottom - box.top
                    box.width = 0
                else:
                    raise Exception("Unexpected box type: %s" % box) # pragma: no cover

            # Line top is the relative offset from the line baseline in canvas coordinates
            line.top = numpy.min([child.top for child in line.children]) if line.children else 0
            # Line bottom is the relative offset from the line baseline in canvas coordinates
            line.bottom = numpy.max([child.bottom for child in line.children]) if line.children else 0

            actual_line_height = line.bottom - line.top
            explicit_line_height = line.style["line-height"]
            offset = (explicit_line_height - actual_line_height) * 0.5
            if offset > 0:
                line.top -= offset
                line.bottom += offset

            line.width = numpy.sum([child.width for child in line.children]) if line.children else 0
            line.height = line.bottom - line.top

        layout.height = numpy.sum([line.height for line in layout.children]) if layout.children else 0


    def compute_position(layout):
        """Compute top + bottom + left + right coordinates for line boxes + text boxes, relative to the layout anchor."""

        if layout.children:
            toyplot_vertical_align = layout.style["-toyplot-vertical-align"]
            # Align the first line's baseline with the anchor.
            if toyplot_vertical_align == "first-baseline":
                offset_y = 0
            # Align the last line's baseline with the anchor.
            elif toyplot_vertical_align == "last-baseline":
                offset_y = -(layout.height + layout.children[0].top - layout.children[-1].bottom)
            # Align the top of the layout with the anchor.
            elif toyplot_vertical_align == "top":
                offset_y = -layout.children[0].top
            # Align the middle of the layout with the anchor.
            elif toyplot_vertical_align == "middle":
                offset_y = -((layout.height * 0.5) + layout.children[0].top)
            # Align the bottom of the layout with the anchor.
            elif toyplot_vertical_align == "bottom":
                offset_y = -(layout.height + layout.children[0].top)
            else:
                raise ValueError("Unknown -toyplot-vertical-align value: %s" % toyplot_vertical_align) # pragma: no cover

            for line in layout.children:
                text_anchor = line.style["text-anchor"] if line.children else "middle"
                if text_anchor == "start":
                    anchor_offset = 0
                elif text_anchor == "middle":
                    anchor_offset = -line.width * 0.5
                elif text_anchor == "end":
                    anchor_offset = -line.width
                else:
                    raise ValueError("Unknown text-anchor value: %s" % text_anchor)
                anchor_offset += layout.style["-toyplot-anchor-shift"]

                offset_x = anchor_offset

                # Line left/right/bottom/top are relative offsets from the layout anchor in canvas coordinates.
                line.left = offset_x
                line.right = offset_x + line.width
                line.top += offset_y
                line.baseline = offset_y
                line.bottom += offset_y

                for child in line.children:
                    # Child left/right/bottom/top are relative offsets from the layout anchor in canvas coordinates.
                    child.left = offset_x
                    child.right = child.left + child.width
                    child.top += offset_y
                    child.baseline += offset_y
                    child.bottom += offset_y
                    # Note that baseline-shift is the opposite of canvas coordinates (positive values shift UP)
                    child.baseline -= child.style["baseline-shift"]

                    offset_x += child.width
                offset_y += line.height

            layout.top = layout.children[0].top
            layout.left = numpy.min([line.left for line in layout.children])
            layout.right = numpy.max([line.right for line in layout.children])
            layout.bottom = layout.children[-1].bottom

        else:
            layout.top = 0
            layout.left = 0
            layout.right = 0
            layout.bottom = 0

        # Layout top/left/right/bottom are relative offsets from the layout anchor in canvas coordinates
        layout.width = layout.right - layout.left
        layout.height = layout.bottom - layout.top


    def cleanup_styles(layout):
        """Remove style properties that we don't want rendered (because their effect is already baked into the box positions."""
        for line in layout.children:
            for child in line.children:
                child.style.pop("-toyplot-anchor-shift", None)
                child.style.pop("-toyplot-vertical-align", None)
                child.style.pop("alignment-baseline", None)
                child.style.pop("baseline-shift", None)
                child.style.pop("text-anchor", None)
                child.style.pop("line-height", None)

    dom = xml.fromstring(("<body>" + text + "</body>").encode("utf-8"))

    default_style = {
        "-toyplot-anchor-shift": "0",
        "-toyplot-vertical-align": "middle",
        "alignment-baseline": "alphabetic",
        "baseline-shift": "0",
        "fill": toyplot.color.black,
        "font-family": "helvetica",
        "font-size": "12px",
        "font-weight": "normal",
        "line-height": "normal",
        "stroke": "none",
        "text-anchor": "middle",
        "vertical-align": "baseline",
        "white-space": "pre",
    }
    style = toyplot.style.combine(default_style, style)
    reference_font_size = toyplot.units.convert(style["font-size"], target="px", default="px")

    cascade_styles(style, dom)
    compute_styles(reference_font_size, dom)

    root = build_formatting_model(dom)
    split_lines(root)
    compute_size(fonts, root)
    compute_position(root)
    cleanup_styles(root)

    return root


def dump(box, stream=None, level=0, indent="  ", recursive=True): # pragma: no cover
    """Writes formatted text layout information to a stream.

    Parameters
    ----------
    box: :class:`toyplot.text.Layout`, :class:`toyplot.text.LineBox`, :class:`toyplot.text.TextBox`, :class:`toyplot.text.MarkerBox`, :class:`toyplot.text.PushHyperlink`, or :class:`toyplot.text.PopHyperlink`, required
        The layout object to be formatted.
    stream: file like object, optional
        The stream where formatted data will be written.  Defaults to
        :data:`sys.stdout`
    recursive: boolean, optional
        If `True`, recursively print the entire layout tree underneath `box`.
    """
    if stream is None:
        stream = sys.stdout
    stream.write(indent * level)
    stream.write("%s.%s\n" % (box.__module__, box.__class__.__name__))

    for key, value in sorted(box.__dict__.items()):
        if key in ["children"]:
            continue
        if key == "style":
            stream.write(indent * (level + 1))
            stream.write("style:\n")
            for pair in sorted(box.style.items()):
                stream.write(indent * (level + 2))
                stream.write("%s: %s\n" % pair)
        elif key in ["text"]:
            stream.write(indent * (level + 1))
            stream.write("%s: %r\n" % (key, value))
        else:
            stream.write(indent * (level + 1))
            stream.write("%s: %s\n" % (key, value))

    stream.write("\n")

    if recursive and hasattr(box, "children"):
        for child in box.children:
            dump(box=child, stream=stream, level=level+1, indent=indent)



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Functions to render PDF documents using ReportLab."""



import io
import reportlab.pdfgen.canvas
import toyplot.reportlab
import toyplot.require
import toyplot.svg


def render(canvas, fobj=None, width=None, height=None, scale=None):
    """Render the PDF representation of a canvas using ReportLab.

    Because the canvas dimensions are specified explicitly at creation time, they
    map directly to real-world units in the output PDF image.  Use one of
    `width`, `height`, or `scale` to override this behavior.

    Parameters
    ----------
    canvas: :class:`toyplot.canvas.Canvas`
      Canvas to be rendered.
    fobj: file-like object or string
      The file to write.  Use a string filepath to write data directly to disk.
      If `None` (the default), the PDF data will be returned to the caller
      instead.
    width: number, string, or (number, string) tuple, optional
      Specify the width of the output image with optional units.  If the units
      aren't specified, defaults to points.  See :ref:`units` for details on
      unit conversion in Toyplot.
    height: number or (number, string) tuple, optional
      Specify the height of the output image with optional units.  If the units
      aren't specified, defaults to points.  See :ref:`units` for details on
      unit conversion in Toyplot.
    scale: number, optional
      Scales the output `canvas` by the given ratio.

    Returns
    -------
    pdf: PDF data, or `None`
      PDF representation of `canvas`, or `None` if the caller specifies the
      `fobj` parameter.

    Examples
    --------

    >>> toyplot.reportlab.pdf.render(canvas, "figure-1.pdf", width=(4, "inches"))
    """
    canvas = toyplot.require.instance(canvas, toyplot.canvas.Canvas)
    svg = toyplot.svg.render(canvas)
    scale = canvas._point_scale(width=width, height=height, scale=scale)
    if fobj is None:
        stream = io.BytesIO()
        surface = reportlab.pdfgen.canvas.Canvas(
            stream, pagesize=(scale * canvas.width, scale * canvas.height))
    else:
        surface = reportlab.pdfgen.canvas.Canvas(
            fobj, pagesize=(scale * canvas.width, scale * canvas.height))
    surface.translate(0, scale * canvas.height)
    surface.scale(1, -1)
    surface.scale(scale, scale)
    toyplot.reportlab.render(svg, surface)
    surface.showPage()
    surface.save()
    if fobj is None:
        return stream.getvalue()



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Classes and functions for projecting coordinates between spaces."""


import custom_inherit
import numpy


def _mix(a, b, amount):
    return ((1.0 - amount) * a) + (amount * b)


def _log(x, base):
    return numpy.log10(numpy.abs(x)) / numpy.log10(base)


def _in_range(a, x, b):
    left = min(a, b)
    right = max(a, b)
    old_settings = numpy.seterr(invalid="ignore")
    result = numpy.logical_and(left <= x, x <= right)
    numpy.seterr(**old_settings)
    return result


class Projection(object, metaclass=custom_inherit.DocInheritMeta(style="numpy_napoleon")):
    """Abstract interface for objects that can map between one-dimensional domain and range spaces.

    .. automethod:: __call__
    """
    def __call__(self, domain_values):
        """Map domain values to range values.

        Parameters
        ----------
        domain_values: :class:`numpy.ndarray` or compatible.
            The domain values to convert.

        Returns
        -------
        range_values: :class:`numpy.ndarray`
            The domain values projected into range values.

        Examples
        --------

        >>> projection = toyplot.projection.linear(0, 1, -1, 1)
        >>> projection(0.75)
        0.5
        """
        raise NotImplementedError() # pragma: no cover

    def inverse(self, range_values):
        """Map range values to domain values.

        Parameters
        ----------
        range_values: :class:`numpy.ndarray` or compatible.
            The range values to convert.

        Returns
        -------
        domain_values: :class:`numpy.ndarray`
            The range values projected into domain values.

        Examples
        --------

        >>> projection = toyplot.projection.linear(0, 1, -1, 1)
        >>> projection.inverse(-0.5)
        0.25
        """
        raise NotImplementedError() # pragma: no cover


class Piecewise(Projection):
    """Projects between domain and range data using a piecewise collection of linear and log segments.

    Parameters
    ----------
    segments: sequence of :class:`toyplot.projection.Piecewise.Segment` objects, required
        Callers must supply one-to-many segments, each of which defines a
        mapping between bounded, non-overlapping regions of the domain
        and range spaces.

    .. automethod:: __call__
    """
    class Segment(object):
        """Defines a mapping between bounded regions of domain-space and range-space.

        Parameters
        ----------
        scale: "linear" or ("log", base) tuple, required

        domain_bounds_min, domain_bounds_max: number, required
            These values define the bounds of this segment in domain space, and
            may include positive or negative infinity.

        domain_min, domain_max: number, required
            These values are mapped to `range_min` and `range_max`, respectively.

        range_min, range_max: number, required
            These values are mapped to `domain_min` and `domain_max`, respectively.

        range_bounds_min, range_bounds_max: number, required
            These values define the bounds of this segment in range space, and
            may include positive or negative infinity.
        """
        class _Container(object):
            pass

        def __init__(
                self,
                scale,
                domain_bounds_min,
                domain_min,
                domain_max,
                domain_bounds_max,
                range_bounds_min,
                range_min,
                range_max,
                range_bounds_max,
            ):
            self.scale = scale
            self.domain = Piecewise.Segment._Container()
            self.domain.bounds = Piecewise.Segment._Container()
            self.domain.bounds.min = domain_bounds_min
            self.domain.min = domain_min
            self.domain.max = domain_max
            self.domain.bounds.max = domain_bounds_max
            self.range = Piecewise.Segment._Container()
            self.range.bounds = Piecewise.Segment._Container()
            self.range.bounds.min = range_bounds_min
            self.range.min = range_min
            self.range.max = range_max
            self.range.bounds.max = range_bounds_max

    def __init__(self, segments):
        self._segments = segments

    def __call__(self, domain_values):
        domain_values = numpy.ma.array(domain_values, dtype="float64")
        range_values = numpy.empty_like(domain_values)
        for segment in self._segments:
            indices = _in_range(
                segment.domain.bounds.min,
                domain_values,
                segment.domain.bounds.max)
            if segment.scale == "linear":
                amount = (domain_values[indices] - segment.domain.min) / (segment.domain.max - segment.domain.min)
                range_values[indices] = _mix(
                    segment.range.min, segment.range.max, amount)
            else:
                scale, base = segment.scale
                if scale == "log":
                    amount = (_log(domain_values[indices],
                                   base) - _log(segment.domain.min,
                                                base)) / (_log(segment.domain.max,
                                                               base) - _log(segment.domain.min,
                                                                            base))
                    range_values[indices] = _mix(
                        segment.range.min, segment.range.max, amount)
                else:
                    raise Exception("Unknown scale: %s" % (scale,)) # pragma: no cover

        if range_values.shape == ():
            range_values = range_values.item()
        return range_values

    def inverse(self, range_values):
        range_values = numpy.ma.array(range_values, dtype="float64")
        domain_values = numpy.empty_like(range_values)
        for segment in self._segments:
            indices = _in_range(
                segment.range.bounds.min,
                range_values,
                segment.range.bounds.max)
            if segment.scale == "linear":
                amount = (range_values[indices] - segment.range.min) / (segment.range.max - segment.range.min)
                domain_values[indices] = _mix(
                    segment.domain.min, segment.domain.max, amount)
            else:
                scale, base = segment.scale
                if scale == "log":
                    amount = (range_values[indices] - segment.range.min) / (segment.range.max - segment.range.min)
                    domain_values[indices] = numpy.sign(
                        segment.domain.min) * numpy.power(base, _mix(
                            _log(segment.domain.min, base),
                            _log(segment.domain.max, base),
                            amount))
                else:
                    raise Exception("Unknown scale: %s" % (scale,)) # pragma: no cover

        if domain_values.shape == ():
            domain_values = domain_values.item()
        return domain_values


def linear(domain_min, domain_max, range_min, range_max):
    """Return an instance of :class:`toyplot.projection.Piecewise` that performs a linear projection.

    Parameters
    -----------
    domain_min, domain_max : number
        Defines a closed interval of domain values that will be mapped to the range.
    range_min, range_max : number
        Defines a closed interval of range values that will be mapped to the domain.

    Returns
    -------
    projection : :class:`toyplot.projection.Piecewise`
    """

    return Piecewise([
        Piecewise.Segment(
            "linear",
            -numpy.inf,
            domain_min,
            domain_max,
            numpy.inf,
            -numpy.inf,
            range_min,
            range_max,
            numpy.inf,
            ),
        ])


def log(
        base,
        domain_min,
        domain_max,
        range_min,
        range_max,
        linear_domain_min=-1,
        linear_domain_max=1,
    ):
    """Return an instance of :class:`toyplot.projection.Piecewise` that performs a log projection.

    The returned projection will work correctly with both positive, negative,
    and zero domain values.  To support mapping zero, the projection will
    switch from log projection to a linear projection within a user-defined
    region around the origin.

    Parameters
    -----------
    base : number
        Logarithmic base used to map from domain values to range values.
    domain_min, domain_max : number
        Defines a closed interval of domain values that will be mapped to the range.
    range_min, range_max : number
        Defines a closed interval of range values that will be mapped to the domain.
    linear_domain_min, linear_domain_max : number, optional
        Defines an interval of domain values around the origin that will be mapped linearly.

    Returns
    -------
    projection : :class:`toyplot.projection.Piecewise`
    """

    # Domain is all positive.
    if 0 < domain_min:
        return Piecewise([
            Piecewise.Segment(
                "linear",
                -numpy.inf,
                domain_min - (domain_max - domain_min),
                domain_min,
                domain_min,
                -numpy.inf,
                range_min - (range_max - range_min),
                range_min,
                range_min,
                ),
            Piecewise.Segment(
                ("log", base),
                domain_min,
                domain_min,
                domain_max,
                numpy.inf,
                range_min,
                range_min,
                range_max,
                numpy.inf,
                ),
            ])

    # Domain is all negative.
    if domain_max < 0:
        return Piecewise([
            Piecewise.Segment(
                ("log", base),
                -numpy.inf,
                domain_min,
                domain_max,
                domain_max,
                -numpy.inf,
                range_min,
                range_max,
                range_max,
                ),
            Piecewise.Segment(
                "linear",
                domain_max,
                domain_max,
                domain_max + (domain_max - domain_min),
                numpy.inf,
                range_max,
                range_max,
                range_max + (range_max - range_min),
                numpy.inf,
                ),
            ])

    # Mixed negative / positive domain
    if domain_min < linear_domain_min and linear_domain_max < domain_max:
        linear_range_min = _mix(range_min, range_max, 0.4)
        linear_range_max = _mix(range_min, range_max, 0.6)
        return Piecewise([
            Piecewise.Segment(
                ("log", base),
                -numpy.inf,
                domain_min,
                linear_domain_min,
                linear_domain_min,
                -numpy.inf,
                range_min,
                linear_range_min,
                linear_range_min,
                ),
            Piecewise.Segment(
                "linear",
                linear_domain_min,
                linear_domain_min,
                linear_domain_max,
                linear_domain_max,
                linear_range_min,
                linear_range_min,
                linear_range_max,
                linear_range_max,
                ),
            Piecewise.Segment(
                ("log", base),
                linear_domain_max,
                linear_domain_max,
                domain_max,
                numpy.inf,
                linear_range_max,
                linear_range_max,
                range_max,
                numpy.inf,
                ),
            ])

    if domain_min < linear_domain_min:
        linear_range_min = _mix(range_min, range_max, 0.8)
        return Piecewise([
            Piecewise.Segment(
                ("log", base),
                -numpy.inf,
                domain_min,
                linear_domain_min,
                linear_domain_min,
                -numpy.inf,
                range_min,
                linear_range_min,
                linear_range_min,
                ),
            Piecewise.Segment(
                "linear",
                linear_domain_min,
                linear_domain_min,
                linear_domain_max,
                numpy.inf,
                linear_range_min,
                linear_range_min,
                range_max,
                numpy.inf,
                ),
            ])

    if linear_domain_max < domain_max:
        linear_range_max = _mix(range_min, range_max, 0.2)
        return Piecewise([
            Piecewise.Segment(
                "linear",
                -numpy.inf,
                domain_min,
                linear_domain_max,
                linear_domain_max,
                -numpy.inf,
                range_min,
                linear_range_max,
                linear_range_max,
                ),
            Piecewise.Segment(
                ("log", base),
                linear_domain_max,
                linear_domain_max,
                domain_max,
                numpy.inf,
                linear_range_max,
                linear_range_max,
                range_max,
                numpy.inf,
                ),
            ])

    return Piecewise([
        Piecewise.Segment(
            "linear",
            -numpy.inf,
            domain_min,
            domain_max,
            numpy.inf,
            -numpy.inf,
            range_min,
            range_max,
            numpy.inf),
        ])



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Functions to render PNG images."""



import toyplot.require
import toyplot.reportlab.png as implementation


def render(canvas, fobj=None, width=None, height=None, scale=None):
    """Render the PNG bitmap representation of a canvas.

    By default, canvas dimensions in CSS pixels are mapped directly to pixels in
    the output PNG image.  Use one of `width`, `height`, or `scale` to override
    this behavior.

    Parameters
    ----------
    canvas: :class:`toyplot.canvas.Canvas`
      Canvas to be rendered.
    fobj: file-like object or string, optional
      The file to write.  Use a string filepath to write data directly to disk.
      If `None` (the default), the PNG data will be returned to the caller
      instead.
    width: number, optional
      Specify the width of the output image in pixels.
    height: number, optional
      Specify the height of the output image in pixels.
    scale: number, optional
      Ratio of output image pixels to `canvas` pixels.

    Returns
    -------
    png: :class:`bytes` containing PNG image data, or `None`
      Returns `None` if the caller specifies the `fobj` parameter, returns the PNG image data otherwise.

    Notes
    -----
    The output PNG is rendered using :func:`toyplot.reportlab.png.render()`.
    This is subject to change.
    """
    canvas = toyplot.require.instance(canvas, toyplot.canvas.Canvas)
    return implementation.render(canvas, fobj, width, height, scale)


def render_frames(canvas, width=None, height=None, scale=None):
    """Render a canvas as a sequence of PNG images.

    By default, canvas dimensions in CSS pixels are mapped directly to pixels in
    the output PNG images.  Use one of `width`, `height`, or `scale` to override
    this behavior.

    Parameters
    ----------
    canvas: :class:`toyplot.canvas.Canvas`
      Canvas to be rendered.
    width: number, optional
      Specify the width of the output image in pixels.
    height: number, optional
      Specify the height of the output image in pixels.
    scale: number, optional
      Ratio of output image pixels to `canvas` pixels.

    Returns
    -------
    frames: Sequence of :class:`bytes` objects containing PNG image data.
      The caller must iterate over the returned frames and is responsible for all
      subsequent processing, including disk I/O, video compression, etc.

    Notes
    -----
    The output PNG images are rendered using
    :func:`toyplot.reportlab.png.render_frames()`.  This is subject to change.

    Examples
    --------
    >>> for frame, png in enumerate(toyplot.png.render_frames(canvas)):
    ...   open("frame-%s.png" % frame, "wb").write(png)
    """
    canvas = toyplot.require.instance(canvas, toyplot.canvas.Canvas)
    return implementation.render_frames(canvas, width, height, scale)



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Generates SVG images.
"""


import xml.etree.ElementTree as xml


import toyplot.html
import toyplot.require


def apply_changes(svg, changes):
    """Modify the SVG DOM representation of a canvas with the given changes."""
    toyplot.html.apply_changes(svg, changes)


def render(canvas, fobj=None, animation=False):
    """Render the SVG representation of a canvas.

    Parameters
    ----------
    canvas: :class:`toyplot.canvas.Canvas`
      The canvas to be rendered.

    fobj: file-like object or string, optional
      The file to write.  Use a string filepath to write data directly to disk.
      If `None` (the default), the SVG tree will be returned to the caller
      instead.

    animation: boolean, optional
      If `True`, return a representation of the changes to be made to the SVG
      tree for animation.

    Returns
    -------
    svg: xml.etree.ElementTree.Element or `None`
      SVG representation of `canvas`, as a DOM tree, or `None` if the caller
      specifies the `fobj` parameter.

    changes: JSON-compatible data structure, or `None`
      JSON-compatible representation of the animated changes to `canvas`.
    """

    canvas = toyplot.require.instance(canvas, toyplot.canvas.Canvas)
    html, html_animation = toyplot.html.render(canvas, animation=True)
    svg = html.find("svg")

    if isinstance(fobj, str):
        with open(fobj, "wb") as stream:
            stream.write(xml.tostring(svg, method="xml"))
    elif fobj is not None:
        fobj.write(xml.tostring(svg, method="xml"))
    else:
        if animation:
            return svg, html_animation
        else:
            return svg



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Bitmap management and manipulation.
"""


import base64
import io
import logging

import numpy
import png

import toyplot.color


log = logging.getLogger(__name__)


def to_png(data, stream, bitdepth=None):
    """Convert an in-memory bitmap to PNG format.

    Parameters
    ----------
    data: :class:`numpy.ndarray`, required
        Source array containing bitmap data to be converted.  Valid array shapes
        are :math:`M \\times N \\times 1` (greyscale data), :math:`M \\times N \\times 2`
        (greyscale plus alpha channel), :math:`M \\times N \\times 3`
        (RGB data), or :math:`M \\times N \\times 4` (RGB + alpha).  Floating
        point values are scaled and converted to unsigned 8 bit integers.
    stream: file-like object, required
        Target file to receive PNG data.
    bitdepth: integer, optional
        Override the default output bit depth.  Allowed color / bitdepth combinations
        are: greyscale (1/2/4/8/16), greyscale + alpha (8/16), RGB (8/16), and RGB + alpha (8/16).
    """
    if data.dtype == toyplot.color.dtype:
        data = numpy.dstack((data["r"], data["g"], data["b"], data["a"]))

    if issubclass(data.dtype.type, numpy.bool_):
        if bitdepth is None:
            bitdepth = 1
    else:
        if bitdepth is None:
            bitdepth = 8

    if issubclass(data.dtype.type, numpy.floating):
        if bitdepth == 8:
            data = (data * 255.0).astype("uint8")
        elif bitdepth == 16:
            data = (data * 65535.0).astype("uint16")

    width = data.shape[1]
    height = data.shape[0]
    greyscale = data.shape[2] < 3
    alpha = data.shape[2] == 2 or data.shape[2] == 4

    writer = png.Writer(width=width, height=height, greyscale=greyscale, alpha=alpha, bitdepth=bitdepth)
    writer.write(stream, numpy.reshape(data, (-1, data.shape[1] * data.shape[2])))


def to_png_data_uri(data):
    """Convert an in-memory bitmap to PNG format, encoded as a data: URI.

    Parameters
    ----------
    data: :class:`numpy.ndarray`, required
        Source array containing bitmap data to be converted.  Valid array shapes
        are :math:`M \\times N \\times 1` (greyscale data), :math:`M \\times N \\times 2`
        (greyscale plus alpha channel), :math:`M \\times N \\times 3`
        (RGB data), or :math:`M \\times N \\times 4` (RGB + alpha).  Floating
        point values are scaled and converted to unsigned 8 bit integers.

    Returns
    -------
    uri: str
        A data: URI containing the base64 encoded bitmap.
    """
    stream = io.BytesIO()
    to_png(data, stream)
    return "data:image/png;base64," + base64.standard_b64encode(stream.getvalue()).decode("ascii")





# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Provides global configuration for Toyplot figures.

You can write to the attributes in :mod:`toyplot.config` if you wish to change
the default behaviors for multiple figures.
"""

import os

autorender = True
"""Default value for the :class:`toyplot.canvas.Canvas` autorender feature."""

autoformat = os.environ.get("TOYPLOT_AUTOFORMAT", "html")
"""Default value for the :class:`toyplot.canvas.Canvas` autoformat feature."""

width = None
"""Default value for the :class:`toyplot.canvas.Canvas` width."""

height = None
"""Default value for the :class:`toyplot.canvas.Canvas` height."""



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Implements the :class:`toyplot.canvas.Canvas` class, which defines the space that is available for creating plots.
"""


import collections
import numbers

import numpy
import toyplot.coordinates
import toyplot.broadcast
import toyplot.color
import toyplot.config
import toyplot.layout
import toyplot.scenegraph
import toyplot.style
import toyplot.units


class AnimationFrame(object):
    """Used to specify modifications to a `toyplot.canvas.Canvas` for animation.

    Do not create AnimationFrame instances yourself, use
    :meth:`toyplot.canvas.Canvas.frame` and
    :meth:`toyplot.canvas.Canvas.frames` instead.
    """
    def __init__(self, changes, count, number, begin, end):
        self._changes = changes
        self._count = count
        self._number = number
        self._begin = begin
        self._end = end

    @property
    def count(self):
        """Total number of frames in the current animation.
        """
        return self._count

    @property
    def number(self):
        """Current animation frame number (zero-based).
        """
        return self._number

    @property
    def begin(self):
        """Beginning of the current frame in seconds.
        """
        return self._begin

    @property
    def end(self):
        """End of the current frame in seconds.
        """
        return self._end

    @property
    def length(self):
        """Duration of the current frame in seconds.
        """
        return self._end - self._begin

    def set_mark_style(self, mark, style):
        """Change the style of a mark.

        Parameters
        ----------
        mark: :class:`toyplot.mark.Mark` instance
        style: dict containing CSS style information
        """
        if not isinstance(mark, toyplot.mark.Mark):
            raise ValueError(
                "Mark style can only be set on toyplot.mark.Mark instances.")

        self._changes.add(self._begin, self._end, "set-mark-style", dict(mark=mark, style=style))

    def set_datum_style(self, mark, series, datum, style):
        """Change the style of one datum in a :class:`toyplot.mark.Mark` at the current frame.

        Parameters
        ----------
        mark: :class:`toyplot.mark.Mark`, required
            Mark containing the datum to modify.
        series: int, required
            Zero-based index of the series containing the datum to modify.
        datum: int, required
            Zero-based index of the datum to modify.
        style: dict, required
            Python dict containing CSS style information to be assigned to the datum.
        """
        if not isinstance(mark, (
                toyplot.mark.BarBoundaries,
                toyplot.mark.BarMagnitudes,
                toyplot.mark.Plot,
                toyplot.mark.Point,
            )):
            raise ValueError("Cannot set datum style for %s." % type(mark))

        self._changes.add(self._begin, self._end, "set-datum-style", dict(mark=mark, series=series, datum=datum, style=style))

    def set_text(self, mark, text, style=None):
        self.set_datum_text(mark, 0, 0, text, style)

    def set_datum_text(self, mark, series, datum, text, style=None):
        """Change the text in a :class:`toyplot.mark.Text` at the current frame.

        Note
        ----
        Currently, animated text completely overwrites the original - thus, it
        cannot inherit style (including color) information, which you will have
        to re-specify explicitly.

        Parameters
        ----------
        mark: :class:`toyplot.mark.Text`, required
            Mark containing the text to modify.
        series: int, required
            Zero-based index of the series containing the datum to modify.
        datum: int, required
            Zero-based index of the datum to modify.
        text: str, required
            String containing the new text to be assigned to the datum.
        style: dict, optional
            Python dict containing CSS style information to be assigned to the datum.
        angle: float, optional
            Angle for the new text.
        """
        if not isinstance(mark, toyplot.mark.Text):
            raise ValueError("mark must be an instance of toyplot.mark.Text.")
        if not isinstance(series, int):
            raise ValueError("series must be an integer index.")
        if not isinstance(datum, int):
            raise ValueError("datum must be an integer index.")
        if not isinstance(text, str):
            raise ValueError("text must be a string.")
        if not isinstance(style, (dict, type(None))):
            raise ValueError("style must be a dict or None.")

        self._changes.add(self._begin, self._end, "set-datum-text", dict(mark=mark, series=series, datum=datum, text=text, style=style))


##########################################################################
# Canvas

class Canvas(object):
    """Top-level container for Toyplot drawings.

    Parameters
    ----------
    width: number, string, or (number, string) tuple, optional
        Width of the canvas.  Assumes CSS pixels if units aren't provided.
        Defaults to 600px (6.25") if unspecified.  See :ref:`units` for details on
        how Toyplot handles real world units.
    height: number, string, or (number, string) tuple, optional
        Height of the canvas.  Assumes CSS pixels if units aren't provided.
        Defaults to the canvas width if unspecified.  See :ref:`units` for details
        on how Toyplot handles real world units.
    style: dict, optional
        Collection of CSS styles to apply to the canvas.
    hyperlink: string, optional
        When specified, the entire canvas is hyperlinked to the given URI.
        Note that hyperlinks set on other entities (such as axes, marks, or
        text) will override this.
    autorender: boolean, optional
        Turn autorendering on / off for this canvas.  Defaults to the value in
        :data:`toyplot.config.autorender`.
    autoformat: string, optional
        Specify the format ("html" or "png") to be used for autorendering this
        canvas.  Defaults to the value in :data:`toyplot.config.autoformat`.

    Examples
    --------

    The following would create a Canvas 8 inches wide and 6 inches tall, with a yellow background:

    >>> canvas = toyplot.Canvas("8in", "6in", style={"background-color":"yellow"})
    """

    class _AnimationChanges(object):
        def __init__(self):
            self._begin = []
            self._end = []
            self._change = []
            self._state = []

        def add(self, begin, end, change, state):
            self._begin.append(begin)
            self._end.append(end)
            self._change.append(change)
            self._state.append(state)

    def __init__(self, width=None, height=None, style=None, hyperlink=None, autorender=None, autoformat=None):
        if width is None:
            width = toyplot.config.width
        if height is None:
            height = toyplot.config.height
        self._width = toyplot.units.convert(width, "px", default="px") if width is not None else 600
        self._height = toyplot.units.convert(height, "px", default="px") if height is not None else self._width
        self._hyperlink = None
        self._style = {
            "background-color": "transparent",
            "border-color": toyplot.color.black,
            "border-style": "none",
            "border-width": 1.0,
            "fill": toyplot.color.black,
            "fill-opacity": 1.0,
            "font-family": "Helvetica",
            "font-size": "12px",
            "opacity": 1.0,
            "stroke": toyplot.color.black,
            "stroke-opacity": 1.0,
            "stroke-width": 1.0,
        }

        self.hyperlink = hyperlink
        self.style = style

        self._changes = toyplot.Canvas._AnimationChanges()
        self._scenegraph = toyplot.scenegraph.SceneGraph()
        self.autorender(autorender, autoformat)

    def _repr_html_(self):
        from . import html
        return toyplot.html.tostring(self, style={"text-align":"center"})

    def _repr_png_(self):
        from . import png
        return toyplot.png.render(self)

    @property
    def height(self):
        """Height of the canvas in CSS pixels."""
        return self._height

    @height.setter
    def height(self, value):
        self._height = toyplot.units.convert(value, "px", default="px")

    @property
    def hyperlink(self):
        """URI that will be hyperlinked from the entire canvas."""
        return self._hyperlink

    @hyperlink.setter
    def hyperlink(self, value):
        self._hyperlink = toyplot.require.hyperlink(value)

    @property
    def style(self):
        """Collection of CSS styles that will be applied to the canvas."""
        return self._style

    @style.setter
    def style(self, value):
        self._style = toyplot.style.combine(
            self._style,
            toyplot.style.require(value, allowed=set(["background-color", "border-color", "border-style", "border-width"])),
            )

    @property
    def width(self):
        """Width of the canvas in CSS pixels."""
        return self._width

    @width.setter
    def width(self, value):
        self._width = toyplot.units.convert(value, "px", default="px")


    def frame(self, begin, end=None, number=None, count=None):
        """Return a single animation frame that can be used to animate the canvas contents.

        Parameters
        ----------
        begin: float
          Specify the frame start time (in seconds).
        end: float, optional
          Specify the frame end time (in seconds).
        number: integer, optional
          Specify a number for this frame.
        count: integer, optional
          Specify the total number of frames of which this frame is one.

        Returns
        -------
        frame: :class:`toyplot.canvas.AnimationFrame` instance.
        """
        if end is None:
            end = begin + (1.0 / 30.0)
        if number is None:
            number = 0
        if count is None:
            count = 1
        return AnimationFrame(changes=self._changes, count=count, number=number, begin=begin, end=end)


    def frames(self, frames):
        """Return a sequence of animation frames that can be used to animate the canvas contents.

        Parameters
        ----------
        frames: integer, tuple, or sequence
          Pass a sequence of values that specify the time (in seconds) of the
          beginning / end of each frame.  Note that the number of frames will be the
          length of the sequence minus one.  Alternatively, you can pass a 2-tuple
          containing the number of frames and the frame rate in frames-per-second.
          Finally, you may simply specify the number of frames, in which case the
          rate will default to 30 frames-per-second.

        Yields
        ------
        frames: :class:`toyplot.canvas.AnimationFrame`
            Use the methods and properties of each frame object to modify the
            state of the canvas over time.
        """
        if isinstance(frames, numbers.Integral):
            frames = (frames, 30.0)

        if isinstance(frames, tuple):
            frame_count, frame_rate = frames
            times = numpy.linspace(0, frame_count / frame_rate, frame_count + 1, endpoint=True)

        for index, (begin, end) in enumerate(zip(times[:-1], times[1:])):
            yield AnimationFrame(changes=self._changes, count=frame_count, number=index, begin=begin, end=end)


    def animation(self):
        """Return a summary of the canvas animation state.

        Returns
        -------
        begin: float
            Start-time of the animation in seconds.
        end: float
            End-time of the animation in seconds.
        changes: object
            Opaque collection of data structures that describe changes to the
            canvas state during animation (if any).
        """

        begin = 0.0
        end = numpy.max(self._changes._end) if self._changes._end else 0.0
        changes = collections.defaultdict(lambda: collections.defaultdict(list))
        for begin, change, state in zip(self._changes._begin, self._changes._change, self._changes._state):
            changes[begin][change].append(state)

        return begin, end, changes


    def autorender(self, enable=None, autoformat=None):
        """Enable / disable canvas autorendering.

        Autorendering - which is enabled by default when a canvas is created -
        controls how the canvas should be displayed automatically without
        caller intervention in certain interactive environments, such as Jupyter
        notebooks.

        Note that autorendering is disabled when a canvas is explicitly
        shown using any of the rendering backends.

        Parameters
        ----------
        enable: boolean, optional
            Turn autorendering on / off for this canvas.  Defaults to the value
            in :data:`toyplot.config.autorender`.
        format: string, optional
            Specify the format ("html" or "png") to be used for autorendering
            this canvas.  Defaults to the value in
            :data:`toyplot.config.autoformat`.
        """
        if enable is None:
            enable = toyplot.config.autorender
        if autoformat is None:
            autoformat = toyplot.config.autoformat

        Canvas._autorender = [entry for entry in Canvas._autorender if entry[0] != self]
        if enable:
            Canvas._autorender.append((self, autoformat))

    def cartesian(
            self,
            aspect=None,
            bounds=None,
            corner=None,
            grid=None,
            hyperlink=None,
            label=None,
            margin=50,
            padding=10,
            palette=None,
            rect=None,
            show=True,
            xlabel=None,
            xmax=None,
            xmin=None,
            xscale="linear",
            xshow=True,
            xticklocator=None,
            ylabel=None,
            ymax=None,
            ymin=None,
            yscale="linear",
            yshow=True,
            yticklocator=None,
        ):
        """Add a set of Cartesian axes to the canvas.

        See Also
        --------
        :ref:`cartesian-coordinates`

        Parameters
        ----------
        aspect: string, optional
            Set to "fit-range" to automatically expand the domain so that its
            aspect ratio matches the aspect ratio of the range.
        bounds: (xmin, xmax, ymin, ymax) tuple, optional
            Use the bounds property to position / size the axes by specifying the
            position of each of its boundaries.  Assumes CSS pixels if units
            aren't provided, and supports all units described in :ref:`units`,
            including percentage of the canvas width / height.
        corner: (corner, inset, width, height) tuple, optional
            Use the corner property to position / size the axes by specifying its
            width and height, plus an inset from a corner of the canvas.  Allowed
            corner values are "top-left", "top", "top-right", "right",
            "bottom-right", "bottom", "bottom-left", and "left".  The width and
            height may be specified using absolute units as described in
            :ref:`units`, or as a percentage of the canvas width / height.  The
            inset only supports absolute drawing units.  All units default to CSS
            pixels if unspecified.
        grid: (rows, columns, index) tuple, or (rows, columns, i, j) tuple, or (rows, columns, i, rowspan, j, columnspan) tuple, optional
            Use the grid property to position / size the axes using a collection of
            grid cells filling the canvas.  Specify the number of rows and columns in
            the grid, then specify either a zero-based cell index (which runs in
            left-ot-right, top-to-bottom order), a pair of i, j cell coordinates, or
            a set of i, column-span, j, row-span coordinates so the legend can cover
            more than one cell.
        hyperlink: string, optional
            When specified, the range (content area of the axes) is hyperlinked
            to the given URI.  Note that this overrides the canvas hyperlink,
            if any, and is overridden by hyperlinks set on other entities such
            as marks or text.
        label: string, optional
            Human-readable axes label.
        margin: number, (top, side) tuple, (top, side, bottom) tuple, or (top, right, bottom, left) tuple, optional
            Specifies the amount of empty space to leave between the axes
            contents and the containing grid cell When using the `grid`
            parameter for positioning.  Assumes CSS pixels by default, and
            supports all of the absolute units described in :ref:`units`.
        padding: number, string, or (number, string) tuple,  optional
            Distance between the axes and plotted data.  Assumes CSS pixels if units aren't provided.
            See :ref:`units` for details on how Toyplot handles real-world units.
        palette: :class:`toyplot.color.Palette`, optional
            Specifies a palette to be used when automatically assigning per-series colors.
        rect: (x, y, width, height) tuple, optional
            Use the rect property to position / size the axes by specifying its
            upper-left-hand corner, width, and height.  Assumes CSS pixels if
            units aren't provided, and supports all units described in
            :ref:`units`, including percentage of the canvas width / height.
        show: bool, optional
            Set to `False` to hide the axes (the axes contents will still be visible).
        xmin, xmax, ymin, ymax: float, optional
            Used to explicitly override the axis domain (normally, the domain is
            implicitly defined by the marks added to the axes).
        xshow, yshow: bool, optional
            Set to `False` to hide individual axes.
        xlabel, ylabel: string, optional
            Human-readable axis labels.
        xticklocator, yticklocator: :class:`toyplot.locator.TickLocator`, optional
            Controls the placement and formatting of axis ticks and tick labels.
        xscale, yscale: "linear", "log", "log10", "log2", or a ("log", <base>) tuple, optional
            Specifies the mapping from data to canvas coordinates along an
            axis.  See :ref:`log-scales` for examples.

        Returns
        -------
        axes: :class:`toyplot.coordinates.Cartesian`
        """
        xmin_range, xmax_range, ymin_range, ymax_range = toyplot.layout.region(
            0, self._width, 0, self._height,
            bounds=bounds,
            rect=rect,
            corner=corner,
            grid=grid,
            margin=margin,
            )

        axes = toyplot.coordinates.Cartesian(
            aspect=aspect,
            hyperlink=hyperlink,
            label=label,
            padding=padding,
            palette=palette,
            scenegraph=self._scenegraph,
            show=show,
            xaxis=None,
            xlabel=xlabel,
            xmax=xmax,
            xmax_range=xmax_range,
            xmin=xmin,
            xmin_range=xmin_range,
            xscale=xscale,
            xshow=xshow,
            xticklocator=xticklocator,
            ylabel=ylabel,
            yaxis=None,
            ymax=ymax,
            ymax_range=ymax_range,
            ymin=ymin,
            ymin_range=ymin_range,
            yscale=yscale,
            yshow=yshow,
            yticklocator=yticklocator,
            )

        self._scenegraph.add_edge(self, "render", axes)

        return axes

    def legend(
            self,
            entries,
            bounds=None,
            corner=None,
            grid=None,
            label=None,
            margin=50,
            rect=None,
            ):
        """Add a legend to the canvas.

        See Also
        --------
        :ref:`labels-and-legends`

        Parameters
        ----------
        entries: sequence of entries to add to the legend Each entry to be
            displayed in the legend must be either a (label, mark) tuple or a
            (label, marker) tuple.  Labels are human-readable text, markers are
            specified using the syntax described in :ref:`markers`, and marks can
            be any instance of :class:`toyplot.mark.Mark`.
        bounds: (xmin, xmax, ymin, ymax) tuple, optional
            Use the bounds property to position / size the legend by specifying the
            position of each of its boundaries.  The boundaries may be specified in
            absolute drawing units, or as a percentage of the canvas width / height
            using strings that end with "%".
        corner: (corner, inset, width, height) tuple, optional
            Use the corner property to position / size the legend by specifying its
            width and height, plus an inset from a corner of the canvas.  Allowed
            corner values are "top-left", "top", "top-right", "right",
            "bottom-right", "bottom", "bottom-left", and "left".  The width and
            height may be specified in absolute drawing units, or as a percentage of
            the canvas width / height using strings that end with "%".  The inset is
            specified in absolute drawing units.
        grid: (rows, columns, index) tuple, or (rows, columns, i, j) tuple, or (rows, columns, i, rowspan, j, columnspan) tuple, optional
            Use the grid property to position / size the legend using a collection of
            grid cells filling the canvas.  Specify the number of rows and columns in
            the grid, then specify either a zero-based cell index (which runs in
            left-ot-right, top-to-bottom order), a pair of i, j cell coordinates, or
            a set of i, column-span, j, row-span coordinates so the legend can cover
            more than one cell.
        label: str, optional
            Optional human-readable label for the legend.
        margin: number, (top, side) tuple, (top, side, bottom) tuple, or (top, right, bottom, left) tuple, optional
            Specifies the amount of empty space to leave between legend and the
            containing grid cell when using the `grid` parameter for positioning.
        rect: (x, y, width, height) tuple, optional
            Use the rect property to position / size the legend by specifying its
            upper-left-hand corner, width, and height.  Each parameter may be specified
            in absolute drawing units, or as a percentage of the canvas width / height
            using strings that end with "%".

        Returns
        -------
        legend: :class:`toyplot.coordinates.Table`
        """

        xmin, xmax, ymin, ymax = toyplot.layout.region(
            0, self._width, 0, self._height,
            bounds=bounds,
            corner=corner,
            grid=grid,
            margin=margin,
            rect=rect,
            )

        table = toyplot.coordinates.Table(
            annotation=True,
            brows=0,
            columns=2,
            filename=None,
            label=label,
            lcolumns=0,
            scenegraph=self._scenegraph,
            rcolumns=0,
            rows=len(entries),
            trows=0,
            xmax_range=xmax,
            xmin_range=xmin,
            ymax_range=ymax,
            ymin_range=ymin,
            )

        table.cells.column[0].align = "right"
        table.cells.column[1].align = "left"

        for index, (label, spec) in enumerate(entries): # pylint: disable=redefined-argument-from-local
            if isinstance(spec, toyplot.mark.Mark):
                markers = spec.markers
            else:
                markers = [toyplot.marker.convert(spec)]
            text = ""
            for marker in markers:
                if text:
                    text = text + " "
                text = text + marker

            table.cells.cell[index, 0].data = text
            table.cells.cell[index, 1].data = label

        self._scenegraph.add_edge(self, "render", table)

        return table

    def matrix(
            self,
            data,
            blabel=None,
            blocator=None,
            bounds=None,
            bshow=None,
            colorshow=False,
            corner=None,
            filename=None,
            grid=None,
            label=None,
            llabel=None,
            llocator=None,
            lshow=None,
            margin=50,
            rect=None,
            rlabel=None,
            rlocator=None,
            rshow=None,
            step=1,
            tlabel=None,
            tlocator=None,
            tshow=None,
        ):
        """Add a matrix visualization to the canvas.

        See Also
        --------
        :ref:`matrix-visualization`

        Parameters
        ----------
        data: matrix, or (matrix, :class:`toyplot.color.Map`) tuple, required
            The given data will be used to populate the visualization, using either the
            given colormap or a default.
        blabel: str, optional
            Human readable label along the bottom of the matrix.
        blocator: :class:`toyplot.locator.TickLocator`, optional
            Supplies column labels along the bottom of the matrix.
        bounds: (xmin, xmax, ymin, ymax) tuple, optional
            Use the bounds property to position / size the legend by specifying the
            position of each of its boundaries.  The boundaries may be specified in
            absolute drawing units, or as a percentage of the canvas width / height
            using strings that end with "%".
        bshow: bool, optional
            Set to `False` to hide column labels along the bottom of the matrix.
        colorshow: bool, optional
            Set to `True` to add a color scale to the visualization.
        corner: (corner, inset, width, height) tuple, optional
            Use the corner property to position / size the legend by specifying its
            width and height, plus an inset from a corner of the canvas.  Allowed
            corner values are "top-left", "top", "top-right", "right",
            "bottom-right", "bottom", "bottom-left", and "left".  The width and
            height may be specified in absolute drawing units, or as a percentage of
            the canvas width / height using strings that end with "%".  The inset is
            specified in absolute drawing units.
        filename: str, optional
            Supplies a default filename for users who choose to download the
            matrix contents.  Note that users and web browsers are free to
            ignore or override this.
        grid: (rows, columns, index) tuple, or (rows, columns, i, j) tuple, or (rows, columns, i, rowspan, j, columnspan) tuple, optional
            Use the grid property to position / size the legend using a collection of
            grid cells filling the canvas.  Specify the number of rows and columns in
            the grid, then specify either a zero-based cell index (which runs in
            left-ot-right, top-to-bottom order), a pair of i, j cell coordinates, or
            a set of i, column-span, j, row-span coordinates so the legend can cover
            more than one cell.
        label: str, optional
            Optional human-readable label for the matrix.
        llabel: str, optional
            Human readable label along the left side of the matrix.
        llocator: :class:`toyplot.locator.TickLocator`, optional
            Supplies row labels along the left side of the matrix.
        lshow: bool, optional
            Set to `False` to hide row labels along the left side of the matrix.
        margin: number, (top, side) tuple, (top, side, bottom) tuple, or (top, right, bottom, left) tuple, optional
            Specifies the amount of empty space to leave between the matrix and
            the containing grid cell when using the `grid` parameter for
            positioning.
        rect: (x, y, width, height) tuple, optional
            Use the rect property to position / size the legend by specifying its
            upper-left-hand corner, width, and height.  Each parameter may be specified
            in absolute drawing units, or as a percentage of the canvas width / height
            using strings that end with "%".
        rlabel: str, optional
            Human readable label along the right side of the matrix.
        rlocator: :class:`toyplot.locator.TickLocator`, optional
            Supplies row labels along the right side of the matrix.
        rshow: bool, optional
            Set to `False` to hide row labels along the right side of the matrix.
        step: integer, optional
            Specifies the number of rows or columns to skip between indices.
            This is useful when the matrix cells become too small relative to
            the index text.
        tlabel: str, optional
            Human readable label along the top of the matrix.
        tlocator: :class:`toyplot.locator.TickLocator`, optional
            Supplies column labels along the top of the matrix.
        tshow: bool, optional
            Set to `False` to hide column labels along the top of the matrix.

        Returns
        -------
        axes: :class:`toyplot.coordinates.Table`
        """
        if isinstance(data, tuple):
            matrix = toyplot.require.scalar_matrix(data[0])
            colormap = toyplot.require.instance(data[1], toyplot.color.Map)
        else:
            matrix = toyplot.require.scalar_matrix(data)
            colormap = toyplot.color.brewer.map("BlueRed", domain_min=matrix.min(), domain_max=matrix.max())

        colors = colormap.colors(matrix)

        xmin_range, xmax_range, ymin_range, ymax_range = toyplot.layout.region(
            0, self._width, 0, self._height, bounds=bounds, rect=rect, corner=corner, grid=grid, margin=margin)

        table = toyplot.coordinates.Table(
            annotation=False,
            brows=2,
            columns=matrix.shape[1],
            filename=filename,
            label=label,
            lcolumns=2,
            scenegraph=self._scenegraph,
            rcolumns=2,
            rows=matrix.shape[0],
            trows=2,
            xmax_range=xmax_range,
            xmin_range=xmin_range,
            ymax_range=ymax_range,
            ymin_range=ymin_range,
            )

        table.top.row[[0, 1]].height = 20
        table.bottom.row[[0, 1]].height = 20
        table.left.column[[0, 1]].width = 20
        table.right.column[[0, 1]].width = 20

        table.left.column[1].align = "right"
        table.right.column[0].align = "left"

        table.cells.column[[0, -1]].lstyle = {"font-weight":"bold"}
        table.cells.row[[0, -1]].lstyle = {"font-weight":"bold"}

        if tlabel is not None:
            cell = table.top.row[0].merge()
            cell.data = tlabel

        if llabel is not None:
            cell = table.left.column[0].merge()
            cell.data = llabel
            cell.angle = 90

        if rlabel is not None:
            cell = table.right.column[1].merge()
            cell.data = rlabel
            cell.angle = 90

        if blabel is not None:
            cell = table.bottom.row[1].merge()
            cell.data = blabel

        if tshow is None:
            tshow = True
        if tshow:
            if tlocator is None:
                tlocator = toyplot.locator.Integer(step=step)
            for j, label, title in zip(*tlocator.ticks(0, matrix.shape[1] - 1)): # pylint: disable=redefined-argument-from-local
                table.top.cell[1, int(j)].data = label
                #table.top.cell[1, j].title = title

        if lshow is None:
            lshow = True
        if lshow:
            if llocator is None:
                llocator = toyplot.locator.Integer(step=step)
            for i, label, title in zip(*llocator.ticks(0, matrix.shape[0] - 1)): # pylint: disable=redefined-argument-from-local
                table.left.cell[int(i), 1].data = label
                #table.left.cell[i, 1].title = title

        if rshow is None and rlocator is not None:
            rshow = True
        if rshow:
            if rlocator is None:
                rlocator = toyplot.locator.Integer(step=step)
            for i, label, title in zip(*rlocator.ticks(0, matrix.shape[0] - 1)): # pylint: disable=redefined-argument-from-local
                table.right.cell[int(i), 0].data = label
                #table.right.cell[i, 0].title = title

        if bshow is None and blocator is not None:
            bshow = True
        if bshow:
            if blocator is None:
                blocator = toyplot.locator.Integer(step=step)
            for j, label, title in zip(*blocator.ticks(0, matrix.shape[1] - 1)): # pylint: disable=redefined-argument-from-local
                table.bottom.cell[0, int(j)].data = label
                #table.bottom.cell[0, j].title = title

        table.body.cells.data = matrix
        table.body.cells.format = toyplot.format.NullFormatter()

        for i in numpy.arange(matrix.shape[0]):
            for j in numpy.arange(matrix.shape[1]):
                cell = table.body.cell[i, j]
                cell.style = {"stroke": "none", "fill": toyplot.color.to_css(colors[i, j])}
                cell.title = "%.6f" % matrix[i, j]

        self._scenegraph.add_edge(self, "render", table)

        if colorshow:
            self.color_scale(
                colormap=colormap,
                x1=xmax_range,
                y1=ymax_range,
                x2=xmax_range,
                y2=ymin_range,
                width=10,
                padding=10,
                show=True,
                label="",
                ticklocator=None,
                scale="linear",
                )

        return table

    def color_scale(
            self,
            colormap,
            x1=None,
            y1=None,
            x2=None,
            y2=None,
            bounds=None,
            corner=None,
            grid=None,
            label=None,
            margin=50,
            max=None,
            min=None,
            offset=None,
            padding=10,
            rect=None,
            scale="linear",
            show=True,
            ticklocator=None,
            width=10,
        ):
        """Add a color scale to the canvas.

        The color scale displays a mapping from scalar values to colors, for
        the given colormap.  Note that the supplied colormap must have an
        explicitly defined domain (specified when the colormap was created),
        otherwise the mapping would be undefined.

        Parameters
        ----------
        colormap: :class:`toyplot.color.Map`, required
          Colormap to be displayed.
        min, max: float, optional
          Used to explicitly override the domain that will be shown.
        show: bool, optional
          Set to `False` to hide the axis (the color bar will still be visible).
        label: string, optional
          Human-readable label placed below the axis.
        ticklocator: :class:`toyplot.locator.TickLocator`, optional
          Controls the placement and formatting of axis ticks and tick labels.
        scale: "linear", "log", "log10", "log2", or a ("log", <base>) tuple, optional
          Specifies the mapping from data to canvas coordinates along an axis.

        Returns
        -------
        axes: :class:`toyplot.coordinates.Numberline`
        """
        axes = self.numberline(
            bounds=bounds,
            corner=corner,
            grid=grid,
            label=label,
            margin=margin,
            max=max,
            min=min,
            padding=padding,
            palette=None,
            rect=rect,
            scale=scale,
            show=show,
            ticklocator=ticklocator,
            x1=x1,
            x2=x2,
            y1=y1,
            y2=y2,
            )

        axes.colormap(
            colormap=colormap,
            width=width,
            offset=offset,
        )

        return axes

    def numberline(
            self,
            x1=None,
            y1=None,
            x2=None,
            y2=None,
            bounds=None,
            corner=None,
            grid=None,
            label=None,
            margin=50,
            max=None,
            min=None,
            padding=None,
            palette=None,
            rect=None,
            scale="linear",
            show=True,
            spacing=None,
            ticklocator=None,
        ):
        """Add a 1D numberline to the canvas.

        Parameters
        ----------
        min, max: float, optional
          Used to explicitly override the numberline domain (normally, the domain is
          implicitly defined by any marks added to the numberline).
        show: bool, optional
          Set to `False` to hide the numberline (the numberline contents will still be visible).
        label: string, optional
          Human-readable label placed below the numberline axis.
        ticklocator: :class:`toyplot.locator.TickLocator`, optional
          Controls the placement and formatting of axis ticks and tick labels.  See :ref:`tick-locators`.
        scale: "linear", "log", "log10", "log2", or a ("log", <base>) tuple, optional
          Specifies the mapping from data to canvas coordinates along the axis.  See :ref:`log-scales`.
        spacing: number, string, or (number, string) tuple,  optional
          Distance between plotted data added to the numberline.  Assumes CSS
          pixels if units aren't provided.  See :ref:`units` for details on how
          Toyplot handles real-world units.
        padding: number, string, or (number, string) tuple,  optional
          Distance between the numberline axis and plotted data.  Assumes CSS
          pixels if units aren't provided.  See :ref:`units` for details on how
          Toyplot handles real-world units.  Defaults to the same value as
          `spacing`.

        Returns
        -------
        axes: :class:`toyplot.coordinates.Cartesian`
        """
        xmin_range, xmax_range, ymin_range, ymax_range = toyplot.layout.region(
            0, self._width, 0, self._height, bounds=bounds, rect=rect, corner=corner, grid=grid, margin=margin)

        if x1 is None:
            x1 = xmin_range
        else:
            x1 = toyplot.units.convert(x1, target="px", default="px", reference=self._width)
            if x1 < 0:
                x1 = self._width + x1

        if y1 is None:
            y1 = 0.5 * (ymin_range + ymax_range)
        else:
            y1 = toyplot.units.convert(y1, target="px", default="px", reference=self._height)
            if y1 < 0:
                y1 = self._height + y1

        if x2 is None:
            x2 = xmax_range
        else:
            x2 = toyplot.units.convert(x2, target="px", default="px", reference=self._width)
            if x2 < 0:
                x2 = self._width + x2

        if y2 is None:
            y2 = 0.5 * (ymin_range + ymax_range)
        else:
            y2 = toyplot.units.convert(y2, target="px", default="px", reference=self._height)
            if y2 < 0:
                y2 = self._height + y2

        if spacing is None:
            spacing = 20

        if padding is None:
            padding = spacing

        axes = toyplot.coordinates.Numberline(
            label=label,
            max=max,
            min=min,
            padding=padding,
            palette=palette,
            scenegraph=self._scenegraph,
            scale=scale,
            show=show,
            spacing=spacing,
            ticklocator=ticklocator,
            x1=x1,
            x2=x2,
            y1=y1,
            y2=y2,
            )

        self._scenegraph.add_edge(self, "render", axes)

        return axes

    def table(
            self,
            data=None,
            rows=None,
            columns=None,
            annotation=False,
            bounds=None,
            brows=None,
            corner=None,
            filename=None,
            grid=None,
            label=None,
            lcolumns=None,
            margin=50,
            rcolumns=None,
            rect=None,
            trows=None,
        ):
        """Add a set of table axes to the canvas.

        Parameters
        ----------

        Returns
        -------
        axes: :class:`toyplot.coordinates.Table`
        """
        if data is not None:
            data = toyplot.data.Table(data)
            rows = data.shape[0] if rows is None else max(rows, data.shape[0])
            columns = data.shape[1] if columns is None else max(
                columns, data.shape[1])
            if trows is None:
                trows = 1
        if rows is None or columns is None: # pragma: no cover
            raise ValueError("You must specify data, or rows and columns.")
        if trows is None:
            trows = 0
        if brows is None:
            brows = 0
        if lcolumns is None:
            lcolumns = 0
        if rcolumns is None:
            rcolumns = 0

        xmin_range, xmax_range, ymin_range, ymax_range = toyplot.layout.region(
            0, self._width, 0, self._height,
            bounds=bounds,
            rect=rect,
            corner=corner,
            grid=grid,
            margin=margin,
            )
        table = toyplot.coordinates.Table(
            annotation=annotation,
            brows=brows,
            columns=columns,
            filename=filename,
            label=label,
            lcolumns=lcolumns,
            scenegraph=self._scenegraph,
            rcolumns=rcolumns,
            rows=rows,
            trows=trows,
            xmax_range=xmax_range,
            xmin_range=xmin_range,
            ymax_range=ymax_range,
            ymin_range=ymin_range,
            )

        if data is not None:
            for j, (key, column) in enumerate(data.items()):
                if trows:
                    table.top.cell[trows - 1, j].data = key
                for i, (value, mask) in enumerate(zip(column, numpy.ma.getmaskarray(column))):
                    if not mask:
                        table.body.cell[i, j].data = value
                if issubclass(column._data.dtype.type, numpy.floating):
                    if trows:
                        table.top.cell[0, j].align = "center"
                    table.body.cell[:, j].format = toyplot.format.FloatFormatter()
                    table.body.cell[:, j].align = "separator"
                elif issubclass(column._data.dtype.type, numpy.character):
                    table.cells.cell[:, j].align = "left"
                elif issubclass(column._data.dtype.type, numpy.integer):
                    table.cells.cell[:, j].align = "right"

            if trows:
                # Format top cells for use as a header
                table.top.cells.lstyle = {"font-weight": "bold"}
                # Enable a single horizontal line between top and body.
                table.cells.grid.hlines[trows] = "single"

        self._scenegraph.add_edge(self, "render", table)

        return table

    def image(
            self,
            data,
            bounds=None,
            corner=None,
            grid=None,
            margin=50,
            rect=None,
        ):
        """Add an image to the canvas.

        Parameters
        ----------
        data: image, or (image, colormap) tuple
        bounds: (xmin, xmax, ymin, ymax) tuple, optional
          Use the bounds property to position / size the image by specifying the
          position of each of its boundaries.  Assumes CSS pixels if units
          aren't provided, and supports all units described in :ref:`units`,
          including percentage of the canvas width / height.
        rect: (x, y, width, height) tuple, optional
          Use the rect property to position / size the image by specifying its
          upper-left-hand corner, width, and height.  Assumes CSS pixels if
          units aren't provided, and supports all units described in
          :ref:`units`, including percentage of the canvas width / height.
        corner: (corner, inset, width, height) tuple, optional
          Use the corner property to position / size the image by specifying its
          width and height, plus an inset from a corner of the canvas.  Allowed
          corner values are "top-left", "top", "top-right", "right",
          "bottom-right", "bottom", "bottom-left", and "left".  The width and
          height may be specified using absolute units as described in
          :ref:`units`, or as a percentage of the canvas width / height.  The
          inset only supports absolute drawing units.  All units default to CSS
          pixels if unspecified.
        grid: (rows, columns, index) tuple, or (rows, columns, i, j) tuple, or (rows, columns, i, rowspan, j, columnspan) tuple, optional
          Use the grid property to position / size the image using a collection of
          grid cells filling the canvas.  Specify the number of rows and columns in
          the grid, then specify either a zero-based cell index (which runs in
          left-ot-right, top-to-bottom order), a pair of i, j cell coordinates, or
          a set of i, column-span, j, row-span coordinates so the legend can cover
          more than one cell.
        margin: size of the margin around grid cells, optional
          Specifies the amount of empty space to leave between grid cells When using the
          `grid` parameter for positioning.  Assumes CSS pixels by default, and supports
          all of the absolute units described in :ref:`units`.
        """
        colormap = None
        if isinstance(data, tuple):
            data, colormap = data
            if not isinstance(colormap, toyplot.color.Map):
                raise ValueError("Expected toyplot.color.Map, received %s." % colormap) # pragma: no cover
            data = numpy.atleast_3d(data)
            if data.shape[2] != 1:
                raise ValueError("Expected an image with one channel.") # pragma: no cover
            data = colormap.colors(data)

        xmin_range, xmax_range, ymin_range, ymax_range = toyplot.layout.region(
            0, self._width, 0, self._height,
            bounds=bounds,
            rect=rect,
            corner=corner,
            grid=grid,
            margin=margin,
            )

        mark = toyplot.mark.Image(
            xmin_range,
            xmax_range,
            ymin_range,
            ymax_range,
            data=data,
            )

        self._scenegraph.add_edge(self, "render", mark)

        return mark

    def text(
            self,
            x,
            y,
            text,
            angle=0.0,
            fill=None,
            opacity=1.0,
            title=None,
            style=None):
        """Add text to the canvas.

        Parameters
        ----------
        x, y: float
          Coordinates of the text anchor in canvas drawing units.  Note that canvas
          Y coordinates increase from top-to-bottom.
        text: string
          The text to be displayed.
        title: string, optional
          Human-readable title for the mark.  The SVG / HTML backends render the
          title as a tooltip.
        style: dict, optional
          Collection of CSS styles to apply to the mark.  See
          :class:`toyplot.mark.Text` for a list of useful styles.

        Returns
        -------
        text: :class:`toyplot.mark.Text`
        """
        table = toyplot.data.Table()
        table["x"] = toyplot.require.scalar_vector(x)
        table["y"] = toyplot.require.scalar_vector(y, table.shape[0])
        table["text"] = toyplot.broadcast.pyobject(text, table.shape[0])
        table["angle"] = toyplot.broadcast.scalar(angle, table.shape[0])
        table["fill"] = toyplot.broadcast.pyobject(fill, table.shape[0])
        table["toyplot:fill"] = toyplot.color.broadcast(
            colors=fill,
            shape=(table.shape[0],),
            default=toyplot.color.black,
            )
        table["opacity"] = toyplot.broadcast.scalar(opacity, table.shape[0])
        table["title"] = toyplot.broadcast.pyobject(title, table.shape[0])
        style = toyplot.style.require(style, allowed=toyplot.style.allowed.text)

        mark = toyplot.mark.Text(
            coordinate_axes=["x", "y"],
            table=table,
            coordinates=["x", "y"],
            text=["text"],
            angle=["angle"],
            fill=["toyplot:fill"],
            opacity=["opacity"],
            title=["title"],
            style=style,
            annotation=True,
            filename=None,
            )

        self._scenegraph.add_edge(self, "render", mark)

        return mark

    def _point_scale(self, width=None, height=None, scale=None):
        """Return a scale factor to convert this canvas to a target width or height in points."""
        if numpy.count_nonzero(
                [width is not None, height is not None, scale is not None]) > 1:
            raise ValueError("Specify only one of width, height, or scale.") # pragma: no cover

        if width is not None:
            scale = toyplot.units.convert(
                width, "pt") / toyplot.units.convert((self._width, "px"), "pt")
        elif height is not None:
            scale = toyplot.units.convert(
                height, "pt") / toyplot.units.convert((self._height, "px"), "pt")
        elif scale is None:
            scale = 1.0
        scale *= 72.0 / 96.0
        return scale

    @staticmethod
    def _ipython_post_execute():  # pragma: no cover
        try:
            import IPython.display
            for canvas, autoformat in Canvas._autorender:
                if autoformat == "html":
                    IPython.display.display_html(canvas)
                elif autoformat == "png":
                    IPython.display.display_png(canvas)
        except:
            pass

    @staticmethod
    def _ipython_register():  # pragma: no cover
        try:
            import IPython
            if IPython.get_ipython():
                IPython.get_ipython().events.register(
                    "post_execute", Canvas._ipython_post_execute)
        except:
            pass

Canvas._autorender = []
Canvas._ipython_register()



# Copyright 2014, Sandia Corporation. Under the terms of Contract
# DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain
# rights in this software.

"""Font management and font metrics.
"""


import custom_inherit
import reportlab.pdfbase.pdfmetrics

import toyplot.units


class Font(object, metaclass=custom_inherit.DocInheritMeta(style="numpy_napoleon")):
    """Abstract interface for objects that return information about a specific combination of typeface and size."""
    @property
    def ascent(self):
        """Font ascent (maximum height above the baseline).

        Returns
        -------
        ascent: :class:`number<numbers.Number>`
            ascent of the font in CSS pixels.
        """
        raise NotImplementedError() # pragma: no cover

    @property
    def descent(self):
        """Font descent (maximum height below the baseline).

        Returns
        -------
        descent: :class:`number<numbers.Number>`
            descent of the font in CSS pixels.
        """
        raise NotImplementedError() # pragma: no cover

    def width(self, string):
        """Return the width of a string if rendered using the font.

        Parameters
        ----------
        string: str, required
            The text to be measured.

        Returns
        -------
        width: :class:`number<numbers.Number>`
            Width of the string in CSS pixels, if rendered using the given font and
            font size.
        """
        raise NotImplementedError() # pragma: no cover


class Library(object, metaclass=custom_inherit.DocInheritMeta(style="numpy_napoleon")):
    """Abstract interface for objects that manage a collection of fonts."""
    def font(self, style):
        """Lookup a font using CSS style information and return a corresponding Font object.

        Parameters
        ----------
        style: dict containing CSS style information

        Returns
        -------
        font: instance of :class:`toyplot.font.Font`
        """
        raise NotImplementedError() # pragma: no cover


class ReportlabFont(Font):
    """Use Reportlab to access the metrics for a font.

    Parameters
    ----------
    font_family: str
        PDF font family to use for measurement.
    font_size: number
        Font size for the measurement.  Defaults to CSS pixel units, and
        supports all toyplot :ref:`units`.
    """
    def __init__(self, family, size):
        self._family = family
        self._size = toyplot.units.convert(size, target="pt", default="px")

        ascent, descent = reportlab.pdfbase.pdfmetrics.getAscentDescent(self._family, self._size)
        self._ascent = toyplot.units.convert(ascent, target="px", default="pt")
        self._descent = toyplot.units.convert(descent, target="px", default="pt")

    def __repr__(self): # pragma: no cover
        return "<toyplot.font.ReportlabFont family=%r size=%r ascent=%r descent=%r>" % (self._family, self._size, self.ascent, self.descent)

    @property
    def ascent(self):
        return self._ascent

    @property
    def descent(self):
        return self._descent

    def width(self, string):
        width = reportlab.pdfbase.pdfmetrics.stringWidth(string, self._family, self._size)
        width = toyplot.units.convert(width, target="px", default="pt")
        return width


class ReportlabLibrary(Library):
    """Use Reportlab to provide information about standard PDF fonts."""
    def __init__(self):
        self._cache = dict()

    def font(self, style):
        """
        Returns
        -------
        font: instance of :class:`toyplot.font.ReportlabFont`
        """
        size = style["font-size"]
        bold = True if style.get("font-weight", "") == "bold" else False
        italic = True if style.get("font-style", "") == "italic" else False

        for font_family in style["font-family"].split(","):
            font_family = font_family.lower()
            if font_family not in ReportlabLibrary.font._substitutions:
                continue

            font_family = ReportlabLibrary.font._substitutions[font_family]
            key = (font_family, bold, italic, size)
            if key not in self._cache:
                family = ReportlabLibrary.font._font_table[(font_family, bold, italic)]
                self._cache[key] = ReportlabFont(family, size)

            return self._cache[key]

        raise ValueError("Unknown font family: %s" % style) # pragma: no cover

    font._substitutions = {
        "courier": "courier",
        "helvetica": "helvetica",
        "monospace": "courier",
        "sans-serif": "helvetica",
        "serif": "times",
        "times": "times",
        }

    font._font_table = {
        ("courier", False, False): "Courier",
        ("courier", True, False): "Courier-Bold",
        ("courier", False, True): "Courier-Oblique",
        ("courier", True, True): "Courier-BoldOblique",
        ("helvetica", False, False): "Helvetica",
        ("helvetica", True, False): "Helvetica-Bold",
        ("helvetica", False, True): "Helvetica-Oblique",
        ("helvetica", True, True): "Helvetica-BoldOblique",
        ("times", False, False): "Times-Roman",
        ("times", True, False): "Times-Bold",
        ("times", False, True): "Times-Italic",
        ("times", True, True): "Times-BoldItalic",
        }
